import logging
import math
from collections import OrderedDict

import lingerext
import numpy as np
import torch
from torch.onnx import is_in_onnx_export
from torch.onnx.symbolic_opset9 import max_pool2d as onnx_syms_max_pool2d
from torch.onnx.symbolic_opset11 import (_prepare_onnx_paddings,
                                         constant_pad_nd, reflection_pad,
                                         replication_pad)

from ..config import config
from ..ops.bmm_int import BmmInt
from ..ops.ops import ModuleIntConfig
from ..ops.ops_names import (LINGER_FUNCINT_BMM_COUNTER,
                             LINGER_IQTENSOR_LAYER_COUNTER,
                             LINGER_MIX_INT8_MANUAL_ROUND_LAYERS, LINGER_MODE,
                             LINGER_OBIT)
from ..quant import Quant
from ..utils import Dump, PlatFormQuant, QuantMode, ScalerBuffer
from .iqtensor import (IQTensor, from_torch_tensor, iqTranspose,
                       platform_to_string, quantlinear)
from .module_self import get_current_module
from .requant import Requant

torch_max = torch.max
torch_transpose = torch.transpose
torch_relu = torch.relu
torch_pad = torch.nn.functional.pad
torch_relu_ = torch.relu_
torch_cat = torch.cat
torch_max_pool2d = torch.max_pool2d
torch_sigmoid = torch.sigmoid
torch_sigmoid_ = torch.sigmoid_

torch_tanh = torch.tanh
torch_tanh_ = torch.tanh_
torch_clamp = torch.clamp
torch_clamp_ = torch.clamp_
torch_dropout = torch.nn.functional.dropout
torch_onnx_export = torch.onnx.export
torch_pack_padded_sequence = torch.nn.utils.rnn.pack_padded_sequence
torch_pad_packed_sequence = torch.nn.utils.rnn.pad_packed_sequence

torch_softmax = torch.softmax
torch_logsoftmax = torch.log_softmax
torch_var = torch.var


def forward_torch_tensor(decision_tensor):
    return type(decision_tensor) == torch.Tensor


def find_sigmoidtable(x_int, sigmoid_table):
    x_int_uint8 = x_int + 256
    y_int = torch.where(x_int >= 0, x_int, x_int_uint8)
    y_int = y_int.reshape(-1)
    for i, ele in enumerate(y_int):
        y_int[i] = sigmoid_table[ele]

    return y_int.reshape(x_int.shape)

def channel_shuffle(x, groups):
    batchsize, num_channels, height, width = x.data.size()
    channels_per_group = num_channels // groups
    x = x.view(batchsize, groups, channels_per_group, height, width)
    x = torch.transpose(x, 1, 2).contiguous()
    x = x.view(batchsize, -1, height, width)
    return x

class channelShuffle(torch.autograd.Function):
    @staticmethod
    def forward(self, input, groups):
        self.save_for_backward(input)
        self.groups = groups
        y = channel_shuffle(input, groups)
        return from_torch_tensor(y, input.scale_data, input.bits)

    @staticmethod
    def backward(self, s):
        x, = self.saved_tensors
        groups = self.groups
        x = x.detach().clone().requires_grad_(True)
        grad = None
        with torch.enable_grad():
            y = channel_shuffle(x, groups)
            grad = torch.autograd.grad(y, x, s)
        return grad[0], None

    @staticmethod
    def symbolic(g, x, groups):
        param_dict = dict()
        input_list = [x, ]
        param_dict['groups_i'] = groups

        return g.op("thinker::ShuffleChannel", *input_list, **param_dict)

def channel_shuffle_quant(*args, **kwargs):
    if forward_torch_tensor(args[0]):
        return channel_shuffle(*args, **kwargs)
    assert isinstance(args[0], IQTensor)
    assert hasattr(args[0], 'scale_data')
    assert hasattr(args[0], 'bits')
    return channelShuffle.apply(*args)

class iqRelu(torch.autograd.Function):
    @staticmethod
    def forward(self, input):
        self.save_for_backward(input)
        y = torch_relu(input)
        return from_torch_tensor(y, input.scale_data, input.bits)

    @staticmethod
    def backward(self, s):
        x, = self.saved_tensors
        x = x.detach().clone().requires_grad_(True)
        grad = None
        with torch.enable_grad():
            y = torch_relu(x)
            grad = torch.autograd.grad(y, x, s)
        return grad

    @staticmethod
    def symbolic(g, x):

        input_list = [x, ]

        return g.op("Relu", *input_list)


def _constant_pad_nd(g, input, padding, value=None):
    mode = "constant"
    pad = _prepare_onnx_paddings(g, input.type().dim(), padding)
    return g.op("Pad", input, pad, value, mode_s=mode)


class iqPad(torch.autograd.Function):
    @staticmethod
    def forward(self, scale_data, input, pad, mode='constant', value=0.):
        self.save_for_backward(input)
        self.pad = pad
        self.mode = mode
        self.value = value
        y = torch_pad(input, pad, mode, value)
        if value == 0:
            return from_torch_tensor(y, input.scale_data, input.bits)
        else:
            return y

    @staticmethod
    def backward(self, s):
        x, = self.saved_tensors
        x = x.detach().clone().requires_grad_(True)
        grad = None
        with torch.enable_grad():
            y = torch_pad(x, self.pad, self.mode, self.value)
            grad = torch.autograd.grad(y, (x,), s)
        return None, grad[0], None, None, None

    @staticmethod
    def symbolic(g, scale_data, input, padding, mode='constant', value=0.):
        if mode == "constant":
            if value == 0:
                value = Quant.quant(torch.tensor(
                    value), bits=8, scale=scale_data)[0].char()
                padding = g.op("Constant", value_t=torch.tensor(
                    padding, dtype=torch.int64))
                return _constant_pad_nd(g, input, padding, value)
            else:
                value = torch.tensor(value)
                padding = g.op("Constant", value_t=torch.tensor(
                    padding, dtype=torch.int64))
                return constant_pad_nd(g, input, padding, value)
        elif mode == "reflect":
            padding = g.op("Constant", value_t=torch.tensor(
                padding, dtype=torch.int64))
            return reflection_pad(g, input, padding)
        elif mode == "replicate":
            padding = g.op("Constant", value_t=torch.tensor(
                padding, dtype=torch.int64))
            return replication_pad(g, input, padding)


class iqMaxPool2d(torch.autograd.Function):
    @staticmethod
    def forward(self, input, kernel_size, stride=(), padding=0, dilation=1, ceil_mode=False):
        # venus limits
        assert input.bits in (
            4, 8), f"in iqMaxPool2d op, input bits only support 4/8 bits, but you have input bits {input.bits}"

        self.save_for_backward(input)
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.ceil_mode = ceil_mode
        y = torch_max_pool2d(input, kernel_size, stride,
                             padding, dilation, ceil_mode)
        return from_torch_tensor(y, input.scale_data, input.bits)

    @staticmethod
    def backward(self, s):
        x, = self.saved_tensors
        x = x.detach().clone().requires_grad_(True)
        kernel_size = self.kernel_size
        stride = self.stride
        padding = self.padding
        dilation = self.dilation
        ceil_mode = self.ceil_mode
        grad = None
        with torch.enable_grad():
            y = torch_max_pool2d(x, kernel_size, stride,
                                 padding, dilation, ceil_mode)
            grad = torch.autograd.grad(y, x, s)
        return grad[0], None, None, None, None, None


iqMaxPool2d.symbolic = onnx_syms_max_pool2d


class iqCat(torch.autograd.Function):
    @staticmethod
    def forward(self, local_scale_o, scale_o, running_o, dim, quant_mode, training, prefix, dump, path, *args):
        tensors = args[0:len(args)//2]
        scale_s = args[len(args)//2:]
        self.save_for_backward(*tensors)
        self.dim = dim
        scale_z_iq = local_scale_o
        momentum = 0.1
        if training:
            running_o.mul_(1-momentum).add_(momentum*(127/scale_z_iq()))
            if quant_mode == QuantMode.QValue:
                scale_log = round(math.log(scale_z_iq(), 2))
                scale_z_iq = math.pow(2, scale_log)
            scale_z_iq = ScalerBuffer(scale_z_iq)
        else:
            assert running_o.data > 0, 'Must at least training one batch'
            if quant_mode == QuantMode.QValue:
                scale_log = round(math.log(127/running_o.data, 2))
                scale_z_iq = math.pow(2, scale_log)
            else:
                scale_z_iq = np.float32((math.pow(2, 8-1)-1) / running_o.data)
            scale_z_iq = ScalerBuffer(scale_z_iq)
            scale_o.fill_(scale_z_iq())
        if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
            list_tensor = []
            for m in tensors:
                list_tensor.append(m.scale_to(scale_z_iq(), training))
            y_float = torch_cat(list_tensor, dim)
        else:
            assert False, "linger only support luna quant."
        if dump:
            name_list = ["input", "outputs"]
            attr_list = [tensors, y_float]
            Dump.dump_file(prefix, ".iqCat.", zip(name_list, attr_list), path)
        return from_torch_tensor(y_float, scale_z_iq(), 8)

    @staticmethod
    def backward(self, s):
        tensors = self.saved_tensors
        tensors = [tensor.detach().clone().requires_grad_(True)
                   for tensor in tensors]
        dim = self.dim
        grad = None
        with torch.enable_grad():
            y = torch_cat(tensors, dim)
            grad = torch.autograd.grad(y, tensors, s)
        ret = [None, None, None, None, None, None, None, None, None]+list(grad)
        l = [None for _ in range(len(tensors))]
        return tuple(ret + l)

    @staticmethod
    def symbolic(g, local_scale_o, scale_o, running_o, dim, quant_mode, training, prefix, dump, path, *args):
        param_dict = {}
        input_list = args[0:len(args)//2]
        for i, value in enumerate(args[len(args)//2:]):
            param_dict['scale_x_'+str(i)+"_f"] = value()
        assert len(input_list) == len(param_dict)
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict['scale_o_f'] = scale_o()
        param_dict['dim_i'] = dim
        param_dict['platform_quant_s'] = platform_quant

        return g.op("thinker::iqCat", *input_list, **param_dict)


def relu(*args, **kwargs):
    if forward_torch_tensor(args[0]):
        return torch_relu(*args, **kwargs)
    assert isinstance(args[0], IQTensor)
    assert hasattr(args[0], 'scale_data')
    assert hasattr(args[0], 'bits')
    return iqRelu.apply(args[0])


def relu_(*args, **kwargs):
    return relu(*args, **kwargs)


def max_pool2d(input, kernel_size, stride=(), padding=0, dilation=1, ceil_mode=False):
    if forward_torch_tensor(input):
        return torch_max_pool2d(input, kernel_size, stride, padding, dilation, ceil_mode)
    assert isinstance(input, IQTensor)
    assert hasattr(input, 'scale_data')
    assert hasattr(input, 'bits')
    return iqMaxPool2d.apply(input, kernel_size, stride, padding, dilation, ceil_mode)


def pad(*args, **kwargs):
    if forward_torch_tensor(args[0]):
        return torch_pad(*args, **kwargs)
    assert isinstance(args[0], IQTensor)
    assert hasattr(args[0], 'scale_data')
    assert hasattr(args[0], 'bits')
    return iqPad.apply(args[0].scale_data, *args)


class iqCatLayer(torch.nn.Module):
    def __init__(self):
        super(iqCatLayer, self).__init__()
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.register_buffer('scale_o', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))

    def forward(self, tensors, local_scale_o, dim, quant_mode=QuantMode.QValue):
        parmater_to_function = []
        for s in tensors:
            parmater_to_function.append(s)
        for s in tensors:
            parmater_to_function.append(ScalerBuffer(s.scale_data))

        scale_o = ScalerBuffer(self.scale_o)
        running_o = ScalerBuffer(self.running_o)
        local_scale_o = ScalerBuffer(local_scale_o)
        z = iqCat.apply(local_scale_o, scale_o, running_o, dim, quant_mode,
                        self.training, self.prefix, self.dump, self.path, *parmater_to_function)
        self.scale_o.fill_(scale_o.data)
        self.running_o.fill_(running_o.data)
        return z

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


def cat(tensors, dim=0, out=None):
    is_iq_cat = True
    for t in tensors:
        if not isinstance(t, IQTensor):
            is_iq_cat = False
            break
    module_self = get_current_module()
    if module_self is None:
        is_iq_cat = False
    if not is_iq_cat:
        return torch_cat(tensors, dim, out=out)
    assert out == None, 'iqtensor not support out param cat for now please make sure the out is None'
    assert tensors[0].bits == 8, 'iqcat only support 8bit'
    quant_mode = getattr(module_self, LINGER_MODE, QuantMode.QValue)
    iname_index = getattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER)
    setattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER, iname_index+1)
    var_name = LINGER_MIX_INT8_MANUAL_ROUND_LAYERS + '_iqcat_'+str(iname_index)
    if not module_self.training and not hasattr(module_self, var_name):
        logging.warning(
            'eval module has iqcat layer while do not match training module')
        return torch_cat(tensors, dim, out=out)
    iq_layer = None
    if hasattr(module_self, var_name):
        iq_layer = getattr(module_self, var_name)
    else:
        iq_layer = iqCatLayer()
        iq_layer.training = module_self.training
        iq_layer = iq_layer.to(tensors[0].device)
        setattr(module_self, var_name, iq_layer)
    scale_z = None
    with torch.no_grad():
        max_z = -1
        for m in tensors:
            max_z_t = torch.max(torch.abs(m)).item()
            if max_z < max_z_t:
                max_z = max_z_t
        if max_z == 0:
            scale_z = 1.0
        else:
            scale_z = 127 / max_z

    return iq_layer(tensors, scale_z, dim, quant_mode)


class softmaxInt(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, dim, data_bits, training, running_x, running_o, scale_x, scale_o, scale_local_o, prefix, dump, path, mode, o_bits, is_not_from_iqtensor):
        momentum = 0.1
        if training:
            ctx.save_for_backward(input)
            ctx.dim = dim
            ctx.o_bits = o_bits
            if isinstance(input, IQTensor):
                q_input, _, max_value_x = Quant().quant(input.data, data_bits, scale_x,
                                                           mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
            else:
                q_input, _, max_value_x = Quant().quant(
                    input.data, data_bits, mode=mode, quant_data='input')
                running_x.mul_(1-momentum).add_(momentum*max_value_x)

            if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                n_dim = len(input.shape)
                # self.dim_ = n_dim - 1 if (self.dim_ == -1) else self.dim_
                dims = [i for i in range(n_dim)]
                dims[dim] = n_dim - 1
                dims[n_dim - 1] = dim
                x_ori = q_input.contiguous()
                if dim != -1 and dim != n_dim - 1:
                    x_ori = x_ori.permute(*dims)
                x_shaped = x_ori.reshape(-1, x_ori.shape[-1])

                l_scale = 25 - int(math.log2(scale_x.data))     # Q25 in
                if l_scale > 0:
                    x_int_shift = (x_shaped * pow(2, l_scale)).int()
                else:
                    x_int_shift = (
                        x_shaped * pow(2, l_scale) + 0.5).floor().int()

                q_output = lingerext.luna_softmax_int(
                    x_int_shift.contiguous(), float(scale_x()))   # Q25->Q15
                q_output.clamp_(0, 2**15-1)
                q_output = q_output.reshape(x_ori.shape)
                if dim != -1 and dim != n_dim - 1:
                    q_output = q_output.permute(*dims)
                scale_local_o.fill_(2**15)
                outputs = Quant().dequant(q_output, scale_local_o)  # Q15->float
            else:
                assert False, 'platform_quant mode donot support for softmaxInt'

            if o_bits is not None:
                q_output, scale_o, max_value_o = Quant().quant(
                    outputs, o_bits, mode=mode, quant_data='output')
                running_o.mul_(1-momentum).add_(momentum*max_value_o)
        else:
            assert running_x > 0, 'invalid running_x = 0, please finetune training before eval'
            if not isinstance(input, IQTensor):
                scale_x = ScalerBuffer(Quant().running_to_scale(running_x, data_bits, mode=mode))
            if o_bits is not None:
                assert running_o > 0, 'invalid running_o = 0 for softmaxInt'
                scale_o = ScalerBuffer(Quant().running_to_scale(running_o, o_bits, mode=mode))
                # scale_o.fill_(2**31)
     
            q_input, _, _ = Quant().quant(input.data, data_bits,
                                             scale_x, mode=mode, quant_data='input')
            if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                n_dim = len(input.shape)
                # self.dim_ = n_dim - 1 if (self.dim_ == -1) else self.dim_
                dims = [i for i in range(n_dim)]
                dims[dim] = n_dim - 1
                dims[n_dim - 1] = dim
                x_ori = q_input.contiguous()
                if dim != -1 and dim != n_dim - 1:
                    x_ori = x_ori.permute(*dims)
                x_shaped = x_ori.reshape(-1, x_ori.shape[-1])

                l_scale = 25 - int(math.log2(scale_x.data))     # Q25 in
                if l_scale > 0:
                    x_int_shift = (x_shaped * pow(2, l_scale)).int()
                else:
                    x_int_shift = (
                        x_shaped * pow(2, l_scale) + 0.5).floor().int()

                q_output = lingerext.luna_softmax_int(
                    x_int_shift.contiguous(), float(scale_x()))   # Q25->Q15
                q_output.clamp_(0, 2**15-1)
                q_output = q_output.reshape(x_ori.shape)
                if dim != -1 and dim != n_dim - 1:
                    q_output = q_output.permute(*dims)
                scale_local_o.fill_(2**15)
                outputs = Quant().dequant(q_output, scale_local_o)  # Q15->float
            else:
                assert False, 'platform_quant mode donot support for softmaxInt'

            if o_bits is not None:
                q_output, _, _ = Quant().quant(outputs, o_bits, scale_o,
                                                  mode=mode, quant_data='output')
                outputs = Quant().dequant(q_output, scale_o)

            if dump:
                name_list = ['input',  'outputs', 'q_input',  'q_outputs']
                attr_list = [input,  outputs, q_input,  q_output]
                Dump.dump_file(prefix, '.SoftmaxInt.',
                               zip(name_list, attr_list), path)

        if o_bits is None:
            return outputs
        elif isinstance(scale_o, float):
            return from_torch_tensor(outputs, scale_o, o_bits)
        elif isinstance(scale_o, torch.Tensor):
            return from_torch_tensor(outputs, scale_o.item(), o_bits)
        else:
            return from_torch_tensor(outputs, scale_o.data, o_bits)

    @staticmethod
    def backward(ctx, s):
        x, = ctx.saved_tensors
        dim = ctx.dim
        x = x.detach().clone().requires_grad_(True)
        grad = None
        with torch.enable_grad():
            y = torch_softmax(x, dim=dim)
            grad = torch.autograd.grad(y, x, s)
        return grad[0], None, None, None, None, None, None, None, None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, input, dim, data_bits, training,
                 running_x, running_o, scale_x, scale_o, scale_local_o,
                 prefix, dump, path, mode, o_bits, is_not_from_iqtensor):
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        if is_not_from_iqtensor:
            op_inner = quantlinear(g, input, scale_x(),
                                   platform_quant, data_bits)
        param_dict = {'scale_x_f': scale_x(
        ),  'data_bits_i': data_bits, 'dim_i': dim}
        input_list = []
        if is_not_from_iqtensor:
            input_list.append(op_inner)
        else:
            input_list.append(input)

        if o_bits is not None:
            param_dict['scale_o_f'] = scale_o()
            param_dict['o_bits_i'] = o_bits
        param_dict['platform_quant_s'] = platform_quant

        return g.op("thinker::SoftmaxInt", *input_list, **param_dict)


class softmaxIntLayer(torch.nn.Module):
    def __init__(self, data_bits=8, mode=QuantMode.QValue, o_bits=8):
        super(softmaxIntLayer, self).__init__()
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.data_bits = data_bits
        self.o_bits = o_bits
        self.mode = mode
        self.is_not_from_iqtensor = True
        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))
        self.register_buffer('scale_local_o', torch.zeros(1))

    def forward(self, input, dim=-1):
        running_x = ScalerBuffer(self.running_x)
        running_o = ScalerBuffer(self.running_o)
        scale_x = ScalerBuffer(self.scale_x)
        scale_o = ScalerBuffer(self.scale_o)
        scale_local_o = ScalerBuffer(self.scale_local_o)

        if isinstance(input, IQTensor):
            self.is_not_from_iqtensor = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)

        z = softmaxInt.apply(input, dim, self.data_bits, self.training, running_x, running_o, scale_x, scale_o, scale_local_o,
                             self.prefix, self.dump, self.path, self.mode, self.o_bits, self.is_not_from_iqtensor)
        self.running_x.fill_(running_x())
        self.running_o.fill_(running_o())
        self.scale_x.fill_(scale_x())
        self.scale_o.fill_(scale_o())

        return z

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        destination._metadata[prefix[:-1]
                              ] = local_metadata = dict(version=self._version)
        if is_in_onnx_export():
            assert self.running_x > 0, 'invalid running_x <=0'
            scale_x = ScalerBuffer(self.scale_x.data)
            if self.is_not_from_iqtensor:
                scale_x = ScalerBuffer(Quant().running_to_scale(
                    self.running_x, self.data_bits, mode=self.mode))
                self.scale_x.data.fill_(scale_x())

            if self.o_bits is not None:
                scale_o = ScalerBuffer(Quant().running_to_scale(
                    self.running_o, self.o_bits, mode=self.mode))
                # scale_o.fill_(2**31)
                self.scale_o.data.fill_(scale_o())
        self._save_to_state_dict(destination, prefix, keep_vars)
        for name, module in self._modules.items():
            if module is not None:
                module.state_dict(destination, prefix +
                                  name + '.', keep_vars=keep_vars)
        for hook in self._state_dict_hooks.values():
            hook_result = hook(self, destination, prefix, local_metadata)
            if hook_result is not None:
                destination = hook_result
        return destination

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


def softmax(tensor, dim, dtype=None):
    is_softmax_int = True
    if not isinstance(tensor, IQTensor):
        is_softmax_int = False
    module_self = get_current_module()
    if module_self is None:
        is_softmax_int = False
    if not is_softmax_int:
        scale_o = 1
        tensor.clamp_(-128, 127) 
        min_x = torch.min(tensor)
        max_x = torch.max(tensor)
        if min_x == max_x == 0:
            scale_o = math.pow(2, 8)
        else:
            max_abs = torch.max(-min_x, max_x)
            max_value = round(math.log((127) / max_abs, 2))
            scale_o = math.pow(2, max_value)
        tensor = from_torch_tensor(tensor, scale_o, 8)
        # return torch_softmax(tensor, dim=dim)
    # assert tensor.bits == 8, 'softmaxInt only support 8bit'
    quant_mode = getattr(module_self, LINGER_MODE, QuantMode.QValue)
    iname_index = getattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER)
    setattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER, iname_index+1)

    var_name = LINGER_MIX_INT8_MANUAL_ROUND_LAYERS + \
        '_SoftmaxInt_'+str(iname_index)
    iq_layer = None
    if hasattr(module_self, var_name):
        iq_layer = getattr(module_self, var_name)
    else:
        iq_layer = softmaxIntLayer(mode=quant_mode)
        iq_layer.training = module_self.training
        iq_layer = iq_layer.to(tensor.device)
        setattr(module_self, var_name, iq_layer)
    # iq_layer.o_bits = 32
    return iq_layer(tensor, dim=dim)


class logsoftmaxInt(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, dim, data_bits, training, running_x, running_o, scale_x, scale_o, scale_local_o, prefix, dump, path, mode, o_bits, is_not_from_iqtensor):
        momentum = 0.1
        if training:
            ctx.save_for_backward(input)
            ctx.dim = dim
            ctx.o_bits = o_bits
            if isinstance(input, IQTensor):
                q_input, _, max_value_x = Quant().quant(input.data, data_bits, scale_x,
                                                           mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
            else:
                q_input, _, max_value_x = Quant().quant(
                    input.data, data_bits, mode=mode, quant_data='input')
                running_x.mul_(1-momentum).add_(momentum*max_value_x)

            if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                n_dim = len(input.shape)
                # self.dim_ = n_dim - 1 if (self.dim_ == -1) else self.dim_
                dims = [i for i in range(n_dim)]
                dims[dim] = n_dim - 1
                dims[n_dim - 1] = dim
                x_ori = q_input.contiguous()
                if dim != -1 and dim != n_dim - 1:
                    x_ori = x_ori.permute(*dims)
                x_shaped = x_ori.reshape(-1, x_ori.shape[-1])

                l_scale = 25 - int(math.log2(scale_x.data))     # Q25 in
                if l_scale > 0:
                    x_int_shift = (x_shaped * pow(2, l_scale)).int()
                else:
                    x_int_shift = (
                        x_shaped * pow(2, l_scale) + 0.5).floor().int()
                q_output = lingerext.luna_logsoftmax_int(
                    x_int_shift.contiguous(), float(scale_x()))   # Q25->Q25
                q_output.clamp_(-2**31, 0)
                q_output = q_output.reshape(x_ori.shape)
                if dim != -1 and dim != n_dim - 1:
                    q_output = q_output.permute(*dims)
                scale_local_o.fill_(2**25)
                outputs = Quant().dequant(q_output, scale_local_o)  # Q25->float
            else:
                assert False, 'platform_quant mode donot support for logsoftmaxInt'

            if o_bits is not None:
                q_output, scale_o, max_value_o = Quant().quant(
                    outputs, o_bits, mode=mode, quant_data='output')
                running_o.mul_(1-momentum).add_(momentum*max_value_o)
        else:
            assert running_x > 0, 'invalid running_x = 0, please finetune training before eval'
            if not isinstance(input, IQTensor):
                scale_x = ScalerBuffer(Quant().running_to_scale(
                    running_x, data_bits, mode=mode))
            if o_bits is not None:
                assert running_o > 0, 'invalid running_o = 0 for logsoftmaxInt'
                scale_o = ScalerBuffer(Quant().running_to_scale(
                    running_o, o_bits, mode=mode))
                if scale_o.data > float(2**31):
                    scale_o.fill_(2**31)
            q_input, _, _ = Quant().quant(input.data, data_bits,
                                             scale_x, mode=mode, quant_data='input')
            if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                n_dim = len(input.shape)
                # self.dim_ = n_dim - 1 if (self.dim_ == -1) else self.dim_
                dims = [i for i in range(n_dim)]
                dims[dim] = n_dim - 1
                dims[n_dim - 1] = dim
                x_ori = q_input.contiguous()
                if dim != -1 and dim != n_dim - 1:
                    x_ori = x_ori.permute(*dims)
                x_shaped = x_ori.reshape(-1, x_ori.shape[-1])

                l_scale = 25 - int(math.log2(scale_x.data))     # Q25 in
                if l_scale > 0:
                    x_int_shift = (x_shaped * pow(2, l_scale)).int()
                else:
                    x_int_shift = (
                        x_shaped * pow(2, l_scale) + 0.5).floor().int()

                q_output = lingerext.luna_logsoftmax_int(
                    x_int_shift.contiguous(), float(scale_x()))   # Q25->Q25
                q_output.clamp_(-2**31, 0)
                q_output = q_output.reshape(x_ori.shape)
                if dim != -1 and dim != n_dim - 1:
                    q_output = q_output.permute(*dims)
                scale_local_o.fill_(2**25)
                outputs = Quant().dequant(q_output, scale_local_o)  # Q25->float
            else:
                assert False, 'platform_quant mode donot support for logsoftmaxInt'

            if o_bits is not None:
                q_output, _, _ = Quant().quant(outputs, o_bits, scale_o,
                                                  mode=mode, quant_data='output')
                outputs = Quant().dequant(q_output, scale_o)

            if dump:
                name_list = ['input',  'outputs', 'q_input',  'q_outputs']
                attr_list = [input,  outputs, q_input,  q_output]
                Dump.dump_file(prefix, '.SoftmaxInt.',
                               zip(name_list, attr_list), path)

        if o_bits is None:
            return outputs
        elif isinstance(scale_o, float):
            return from_torch_tensor(outputs, scale_o, o_bits)
        elif isinstance(scale_o, torch.Tensor):
            return from_torch_tensor(outputs, scale_o.item(), o_bits)
        else:
            return from_torch_tensor(outputs, scale_o.data, o_bits)

    @staticmethod
    def backward(ctx, s):
        x, = ctx.saved_tensors
        dim = ctx.dim
        x = x.detach().clone().requires_grad_(True)
        grad = None
        with torch.enable_grad():
            y = torch_logsoftmax(x, dim=dim)
            grad = torch.autograd.grad(y, x, s)
        return grad[0], None, None, None, None, None, None, None, None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, input, dim, data_bits, training,
                 running_x, running_o, scale_x, scale_o, scale_local_o,
                 prefix, dump, path, mode, o_bits, is_not_from_iqtensor):
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        if is_not_from_iqtensor:
            op_inner = quantlinear(g, input, scale_x(),
                                   platform_quant, data_bits)
        param_dict = {'scale_x_f': scale_x(
        ),  'data_bits_i': data_bits, 'dim_i': dim}
        input_list = []
        if is_not_from_iqtensor:
            input_list.append(op_inner)
        else:
            input_list.append(input)
        if o_bits is not None:
            param_dict['scale_o_f'] = scale_o()
            param_dict['o_bits_i'] = o_bits
        param_dict['platform_quant_s'] = platform_quant
        return g.op("thinker::LogSoftmaxInt", *input_list, **param_dict)


class logsoftmaxIntLayer(torch.nn.Module):
    def __init__(self, data_bits=8, mode=QuantMode.QValue, o_bits=8):
        super(logsoftmaxIntLayer, self).__init__()
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.data_bits = data_bits
        self.o_bits = o_bits
        self.mode = mode
        self.is_not_from_iqtensor = True
        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))
        self.register_buffer('scale_local_o', torch.zeros(1))

    def forward(self, input, dim=-1):
        running_x = ScalerBuffer(self.running_x)
        running_o = ScalerBuffer(self.running_o)
        scale_x = ScalerBuffer(self.scale_x)
        scale_o = ScalerBuffer(self.scale_o)
        scale_local_o = ScalerBuffer(self.scale_local_o)

        if isinstance(input, IQTensor):
            self.is_not_from_iqtensor = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)

        z = logsoftmaxInt.apply(input, dim, self.data_bits, self.training, running_x, running_o, scale_x, scale_o, scale_local_o,
                                self.prefix, self.dump, self.path, self.mode, self.o_bits, self.is_not_from_iqtensor)
        self.running_x.fill_(running_x())
        self.running_o.fill_(running_o())
        self.scale_x.fill_(scale_x())
        self.scale_o.fill_(scale_o())

        return z

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        destination._metadata[prefix[:-1]
                              ] = local_metadata = dict(version=self._version)
        if is_in_onnx_export():
            assert self.running_x > 0, 'invalid running_x <=0'
            scale_x = ScalerBuffer(self.scale_x.data)
            if self.is_not_from_iqtensor:
                scale_x = ScalerBuffer(Quant().running_to_scale(
                    self.running_x, self.data_bits, mode=self.mode))
                self.scale_x.data.fill_(scale_x())
            if self.o_bits is not None:
                scale_o = ScalerBuffer(Quant().running_to_scale(
                    self.running_o, self.o_bits, mode=self.mode))
                if scale_o.data > float(2**31):
                    scale_o.fill_(2**31)
                self.scale_o.data.fill_(scale_o())
        self._save_to_state_dict(destination, prefix, keep_vars)
        for name, module in self._modules.items():
            if module is not None:
                module.state_dict(destination, prefix +
                                  name + '.', keep_vars=keep_vars)
        for hook in self._state_dict_hooks.values():
            hook_result = hook(self, destination, prefix, local_metadata)
            if hook_result is not None:
                destination = hook_result
        return destination

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


def logsoftmax(tensor, dim, dtype=None):
    is_logsoftmax_int = True
    if not isinstance(tensor, IQTensor):
        is_logsoftmax_int = False
    module_self = get_current_module()
    if module_self is None:
        is_logsoftmax_int = False
    if not is_logsoftmax_int:
        return torch_logsoftmax(tensor, dim=dim)
    # assert tensor.bits == 8, 'LogsoftmaxInt only support 8bit'
    quant_mode = getattr(module_self, LINGER_MODE, QuantMode.QValue)
    iname_index = getattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER)
    setattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER, iname_index+1)

    var_name = LINGER_MIX_INT8_MANUAL_ROUND_LAYERS + \
        '_LogSoftmaxInt_'+str(iname_index)
    iq_layer = None
    if hasattr(module_self, var_name):
        iq_layer = getattr(module_self, var_name)
    else:
        iq_layer = logsoftmaxIntLayer(mode=quant_mode)
        iq_layer.training = module_self.training
        iq_layer = iq_layer.to(tensor.device)
        setattr(module_self, var_name, iq_layer)
    # iq_layer.o_bits = 32

    return iq_layer(tensor, dim=dim)


class iqSigmoid(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, scale_x, local_scale_o, running_o, scale_o, training, quant_mode, prefix, dump, path, data_bits):
        ctx.save_for_backward(x)
        # x_int = x.quant_to_int8(scale_x())
        # x_int = x_int.contiguous().int()
        scale_z_iq = local_scale_o
        momentum = 0.1
        bound_value = 127

        if training:
            running_o.mul_(1-momentum).add_(momentum *
                                            (bound_value/local_scale_o()))
            if quant_mode == QuantMode.QValue:
                scale_log = round(math.log(scale_z_iq(), 2))
                scale_z_iq = math.pow(2, scale_log)
            scale_z_iq = ScalerBuffer(scale_z_iq)
        else:
            assert running_o.data > 0, 'Must at least training one batch'
            if quant_mode == QuantMode.QValue:
                scale_log = round(math.log(bound_value/running_o.data, 2))
                scale_z_iq = math.pow(2, scale_log)
            else:
                scale_z_iq = np.float32(bound_value / running_o.data)
            scale_z_iq = ScalerBuffer(scale_z_iq)
            scale_o.fill_(scale_z_iq())
        y_int = None

        if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
            q_input, _, _ = Quant().quant(
                x.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=x.zero_point)
            l_scale = 11 - int(math.log2(scale_x.data))

            if l_scale > 0:
                x_int_shift = (q_input * pow(2, l_scale)).int()
            else:
                x_int_shift = (q_input * pow(2, l_scale) + 0.5).floor().int()

            y_int = lingerext.luna_iqsigmoid(
                x_int_shift.contiguous(), float(scale_x()))
            y_int.clamp_(0, 2**7-1)
            scale_z_iq.fill_(2**7)
            scale_o.fill_(scale_z_iq())
            running_o.fill_(1.0)
            y_float = Quant.dequant(y_int, scale_z_iq)

            if dump:
                name_list = ['input',  'outputs', 'q_input',  'q_outputs']
                attr_list = [x,  y_float, q_input,  y_int]
                Dump.dump_file(prefix, '.iqSigmoid.',
                                zip(name_list, attr_list), path)

            return from_torch_tensor(y_float, scale_z_iq(), 8, zero_point=0)
        else:
            assert False, 'platform_quant mode donot support for iqSigmoid'

    @staticmethod
    def backward(ctx, s):
        x, = ctx.saved_tensors
        x = x.detach().clone().requires_grad_(True)
        grad = None
        with torch.enable_grad():
            y = torch_sigmoid(x)
            grad = torch.autograd.grad(y, x, s)
        return grad[0], None, None, None, None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, x, scale_x, local_scale_o, running_o, scale_o, training, quant_mode, prefix, dump, path, data_bits):
        param_dict = {'scale_x_f': scale_x(), 'scale_o_f': scale_o()}
        input_list = [x, ]
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict['platform_quant_s'] = platform_quant
        param_dict['castor_mode_s'] = "luna"
        op = None
        op = g.op("thinker::iqSigmoid", *input_list, **param_dict)
        return op


class iqSigmoidLayer(torch.nn.Module):
    def __init__(self, data_bits=16):
        super(iqSigmoidLayer, self).__init__()
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.data_bits = data_bits
        self.register_buffer('scale_o', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))

    def forward(self, input, local_scale_o, quant_mode=QuantMode.QValue):
        scale_x = ScalerBuffer(input.scale_data)
        local_scale_o = ScalerBuffer(local_scale_o)
        scale_o = ScalerBuffer(self.scale_o)
        running_o = ScalerBuffer(self.running_o)
        if isinstance(input, IQTensor):
            input = Requant.apply(
                input, input.bits, input.scale_data, self.data_bits)
            scale_x = ScalerBuffer(input.scale_data)
        z = iqSigmoid.apply(input, scale_x, local_scale_o, running_o, scale_o,
                            self.training, quant_mode, self.prefix, self.dump, self.path, self.data_bits)
        self.scale_o.fill_(scale_o.data)
        self.running_o.fill_(running_o.data)
        return z
    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


def sigmoid(tensor, *, out=None):
    is_iq_sigmoid = True
    if not isinstance(tensor, IQTensor):
        is_iq_sigmoid = False
    module_self = get_current_module()
    if module_self is None:
        is_iq_sigmoid = False
    if not is_iq_sigmoid:
        return torch_sigmoid(tensor)
    # assert tensor.bits == 8, 'iqsigmoid only support 8bit'
    quant_mode = getattr(module_self, LINGER_MODE, QuantMode.QValue)
    iname_index = getattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER)
    setattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER, iname_index+1)

    var_name = LINGER_MIX_INT8_MANUAL_ROUND_LAYERS + \
        '_iqsigmoid_'+str(iname_index)
    iq_layer = None
    if hasattr(module_self, var_name):
        iq_layer = getattr(module_self, var_name)
    else:
        iq_layer = iqSigmoidLayer()
        iq_layer.training = module_self.training
        iq_layer = iq_layer.to(tensor.device)
        setattr(module_self, var_name, iq_layer)
    scale_z = None
    with torch.no_grad():
        z_f = torch_sigmoid(tensor)
        max_z = torch.max(torch.abs(z_f))
        if max_z == 0:
            scale_z = 1.0
        else:
            scale_z = 127 / max_z.item()
    return iq_layer(tensor, scale_z, quant_mode)


def sigmoid_(tensor, *, out=None):
    return sigmoid(tensor, out)


class iqTanh(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, scale_x, local_scale_o, running_o, scale_o, training, quant_mode, prefix, dump, path, data_bits):
        ctx.save_for_backward(x)
        # x_int = x.quant_to_int8(scale_x())
        # x_int = x_int.contiguous().int()
        scale_z_iq = local_scale_o
        momentum = 0.1
        bound_value = 127

        if training:
            running_o.mul_(1-momentum).add_(momentum *
                                            (bound_value/local_scale_o()))
            if quant_mode == QuantMode.QValue:
                scale_log = round(math.log(scale_z_iq(), 2))
                scale_z_iq = math.pow(2, scale_log)
            scale_z_iq = ScalerBuffer(scale_z_iq)
        else:
            assert running_o.data > 0, 'Must at least training one batch'
            if quant_mode == QuantMode.QValue:
                scale_log = round(math.log(bound_value/running_o.data, 2))
                scale_z_iq = math.pow(2, scale_log)
            else:
                scale_z_iq = np.float32(bound_value / running_o.data)
            scale_z_iq = ScalerBuffer(scale_z_iq)
            scale_o.fill_(scale_z_iq())
        y_int = None

        if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
            q_input, _, _ = Quant().quant(
                x.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=x.zero_point)
            l_scale = 11 - int(math.log2(scale_x.data))

            if l_scale > 0:
                x_int = (q_input * pow(2, l_scale)).int()
            else:
                x_int = (q_input * pow(2, l_scale) + 0.5).floor().int()
            y_int = lingerext.luna_iqtanh(x_int.contiguous(), float(scale_x()))
            scale_z_iq.fill_(2**7)
            scale_o.fill_(scale_z_iq())
            running_o.fill_(1.0)
            y_float = Quant.dequant(y_int, scale_z_iq)
            return from_torch_tensor(y_float, scale_z_iq(), 8, zero_point=0)

        else:
            assert False, "linger only support luna quant."

    @staticmethod
    def backward(ctx, s):
        x, = ctx.saved_tensors
        x = x.detach().clone().requires_grad_(True)
        grad = None
        with torch.enable_grad():
            y = torch_tanh(x)
            grad = torch.autograd.grad(y, x, s)
        return grad[0], None, None, None, None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, x, scale_x, local_scale_o, running_o, scale_o, training, quant_mode, prefix, dump, path, data_bits):
        param_dict = {'scale_x_f': scale_x(), 'scale_o_f': scale_o()}
        input_list = [x, ]
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict['platform_quant_s'] = platform_quant
        param_dict['castor_mode_s'] = "luna"
        op = None
        op = g.op("thinker::iqTanh", *input_list, **param_dict)
        return op


class iqTanhLayer(torch.nn.Module):
    def __init__(self, data_bits=16):
        super(iqTanhLayer, self).__init__()
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.data_bits = data_bits
        self.register_buffer('scale_o', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))

    def forward(self, input, local_scale_o, quant_mode=QuantMode.QValue):
        scale_x = ScalerBuffer(input.scale_data)
        local_scale_o = ScalerBuffer(local_scale_o)
        scale_o = ScalerBuffer(self.scale_o)
        running_o = ScalerBuffer(self.running_o)
        if isinstance(input, IQTensor):
            input = Requant.apply(
                input, input.bits, input.scale_data, self.data_bits)
            scale_x = ScalerBuffer(input.scale_data)
        z = iqTanh.apply(input, scale_x, local_scale_o, running_o, scale_o,
                         self.training, quant_mode, self.prefix, self.dump, self.path, self.data_bits)
        self.scale_o.fill_(scale_o.data)
        self.running_o.fill_(running_o.data)
        return z
    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


def tanh(tensor, *, out=None):
    is_iq_tanh = True
    if not isinstance(tensor, IQTensor):
        is_iq_tanh = False
    module_self = get_current_module()
    if module_self is None:
        is_iq_tanh = False
    if not is_iq_tanh:
        return torch_tanh(tensor)
    # assert tensor.bits == 8, 'iqTanh only support 8bit'
    quant_mode = getattr(module_self, LINGER_MODE, QuantMode.QValue)
    iname_index = getattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER)
    setattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER, iname_index+1)

    var_name = LINGER_MIX_INT8_MANUAL_ROUND_LAYERS + \
        '_iqtanh_'+str(iname_index)
    iq_layer = None
    if hasattr(module_self, var_name):
        iq_layer = getattr(module_self, var_name)
    else:
        iq_layer = iqTanhLayer()
        iq_layer.training = module_self.training
        iq_layer = iq_layer.to(tensor.device)
        setattr(module_self, var_name, iq_layer)
    scale_z = None
    with torch.no_grad():
        z_f = torch_tanh(tensor)
        max_z = torch.max(torch.abs(z_f))
        if max_z == 0:
            scale_z = 1.0
        else:
            scale_z = 127 / max_z.item()

    return iq_layer(tensor, scale_z, quant_mode)


def tanh_(tensor, *, out=None):
    return tanh(tensor, out)


class iqClamp(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, min, max, scale_x, local_scale_o, running_o, scale_o, training, quant_mode, prefix, dump, path):
        ctx.save_for_backward(x)
        ctx.min = min
        ctx.max = max
        scale_z_iq = local_scale_o
        momentum = 0.1
        if training:
            running_o.mul_(1-momentum).add_(momentum*(127/local_scale_o()))
            if quant_mode == QuantMode.QValue:
                scale_log = round(math.log(scale_z_iq(), 2))
                scale_z_iq = math.pow(2, scale_log)
            scale_z_iq = ScalerBuffer(scale_z_iq)
        else:
            assert running_o.data > 0, 'Must at least training one batch'
            if quant_mode == QuantMode.QValue:
                scale_log = round(math.log(127/running_o.data, 2))
                scale_z_iq = math.pow(2, scale_log)
            else:
                scale_z_iq = np.float32(127 / running_o.data)
            scale_z_iq = ScalerBuffer(scale_z_iq)
            scale_o.fill_(scale_z_iq())
        y_int = None
        y_float = torch_clamp(x, min, max)
        if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
            Qx = math.log(127/max, 2)
            assert Qx.is_integer(
            ), "luna_quant max value don't support {} clamp, it must be (127 /2^n).".format(max)
            assert (min == -128/2**math.log(127/7.9375, 2)) or (min ==
                                                                0), "luna_quant min value don't support {} clamp, it must match with the max-value (-128/2^n) or 0.".format(min)
            scale_z_iq = 2**Qx
            scale_z_iq = ScalerBuffer(scale_z_iq)
            y_int = (x * scale_z_iq + 0.5).floor().int()
            if min == 0:
                y_int = torch_clamp(y_int, 0, 127)
            scale_o.fill_(scale_z_iq())

        else:
            assert False, 'platform_quant mode donot support for iqClamp'
        y_int = torch_clamp(y_int, -128, 127)
        y_float = Quant.dequant(y_int, scale_z_iq)
        if dump:
            name_list = ["input", "outputs", "q_outputs"]
            attr_list = [x, y_float, y_int]
            Dump.dump_file(prefix, ".iqClamp.", zip(
                name_list, attr_list), path)
        return from_torch_tensor(y_float, scale_z_iq(), 8)

    @staticmethod
    def backward(ctx, s):
        x, = ctx.saved_tensors
        min = ctx.min
        max = ctx.max
        x = x.detach().clone().requires_grad_(True)
        grad = None
        with torch.enable_grad():
            y = torch_clamp(x, min, max)
            grad = torch.autograd.grad(y, (x,), s)
        return grad[0], None, None, None, None, None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, x, min, max, scale_x, local_scale_o, running_o, scale_o, training, quant_mode, prefix, dump, path):
        param_dict = {'min_f': min, 'max_f': max,
                      'scale_x_f': scale_x(), 'scale_o_f': scale_o()}
        input_list = [x, ]
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict['platform_quant_s'] = platform_quant

        return g.op("thinker::iqClamp", *input_list, **param_dict)


class iqClampLayer(torch.nn.Module):
    def __init__(self):
        super(iqClampLayer, self).__init__()
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.register_buffer('scale_o', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))

    def forward(self, input, min, max, local_scale_o, quant_mode=QuantMode.QValue):
        scale_x = ScalerBuffer(input.scale_data)
        local_scale_o = ScalerBuffer(local_scale_o)
        scale_o = ScalerBuffer(self.scale_o)
        running_o = ScalerBuffer(self.running_o)
        z = iqClamp.apply(input, min, max, scale_x, local_scale_o, running_o,
                          scale_o, self.training, quant_mode, self.prefix, self.dump, self.path)
        self.scale_o.fill_(scale_o.data)
        self.running_o.fill_(running_o.data)
        return z
    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)


def clamp(tensor, min=-math.inf, max=math.inf, out=None):
    is_iq_clamp = True
    if not isinstance(tensor, IQTensor):
        is_iq_clamp = False
    module_self = get_current_module()
    if module_self is None:
        is_iq_clamp = False
    if not is_iq_clamp:
        return torch_clamp(tensor, min, max)
    assert tensor.bits == 8, 'iqclamp only support 8bit'
    quant_mode = getattr(module_self, LINGER_MODE, QuantMode.QValue)
    iname_index = getattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER)
    setattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER, iname_index+1)
    var_name = LINGER_MIX_INT8_MANUAL_ROUND_LAYERS + \
        '_iqclamp_'+str(iname_index)
    iq_layer = None
    if hasattr(module_self, var_name):
        iq_layer = getattr(module_self, var_name)
    else:
        iq_layer = iqClampLayer()
        iq_layer.training = module_self.training
        iq_layer = iq_layer.to(tensor.device)
        setattr(module_self, var_name, iq_layer)
    scale_z = None
    with torch.no_grad():
        z_f = torch_clamp(tensor, min, max)
        max_z = torch.max(torch.abs(z_f))
        if max_z == 0:
            scale_z = 1.0
        else:
            scale_z = 127 / max_z.item()
    return iq_layer(tensor, min, max, scale_z, quant_mode)


def clamp_(tensor, min=-math.inf, max=math.inf, out=None):
    return clamp(tensor, min, max, out)


def dropout(tensor, p: float, training: bool = False, inplace: bool = False):
    return torch_dropout(tensor, p, False, inplace)


def pack_padded_sequence(input, lengths, batch_first=False, enforce_sorted=True):
    return input, lengths, batch_first, enforce_sorted


def pad_packed_sequence(sequence, batch_first=False, padding_value=0.0, total_length=None):
    assert (padding_value == 0.0 and total_length is None), 'lstmint for pad_packed only support padding_value=0.0 and total_length=None'
    output, lengths = sequence
    return output, lengths


def transpose(*args, **kwargs):
    if forward_torch_tensor(args[0]):
        return torch_transpose(*args, **kwargs)
    assert isinstance(args[0], IQTensor)
    assert hasattr(args[0], 'scale_data')
    assert hasattr(args[0], 'bits')
    return iqTranspose.apply(args[0], args[1], args[2])


class iqMax(torch.autograd.Function):
    @staticmethod
    def forward(self, input, other, scale_x, scale_y, scale_o):
        self.save_for_backward(*(input, other))
        y = torch_max(input, other)

        return from_torch_tensor(y, scale_o, input.bits, input.zero_point)

    @staticmethod
    def backward(self, s):
        x, y = self.saved_tensors
        x = x.detach().clone().requires_grad_(True)
        y = y.detach().clone().requires_grad_(True)

        grad = None
        with torch.enable_grad():
            z = torch_max(x, y)
            grad = torch.autograd.grad(z, (x, y), s)
        return grad[0], grad[1], None, None, None

    @staticmethod
    def symbolic(g, input, other, scale_x, scale_y, scale_o):
        # torch.max(input, other)
        param_dict = {'scale_x_f': scale_x,
                      'scale_y_f': scale_y, 'scale_o_f': scale_o}
        input_list = [input, other, ]
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict['platform_quant_s'] = platform_quant

        op = g.op("thinker::iqMax", *input_list, **param_dict)
        return op


def iqmax(*args, **kwargs):
    if len(args) != 2:
        return torch_max(*args, **kwargs)
    if not isinstance(args[0], IQTensor):
        return torch_max(*args, **kwargs)
    if not isinstance(args[1], IQTensor):
        return torch_max(*args, **kwargs)
    if len(kwargs) != 0:
        return torch_max(*args, **kwargs)
    assert isinstance(args[0], IQTensor)
    assert isinstance(args[1], IQTensor)

    assert hasattr(args[0], 'scale_data')
    assert hasattr(args[0], 'bits')
    assert hasattr(args[1], 'scale_data')
    assert hasattr(args[1], 'bits')

    max_scale_o = min(args[0].scale_data, args[1].scale_data)
    return iqMax.apply(args[0], args[1], args[0].scale_data, args[1].scale_data, max_scale_o)


def bmm(input, mat2, *, out=None):
    module_self = get_current_module()

    quant_mode = getattr(module_self, LINGER_MODE, QuantMode.QValue)

    out_bits = getattr(module_self, LINGER_OBIT,
                       None) if True else None
    iname_index = getattr(module_self, LINGER_FUNCINT_BMM_COUNTER)
    setattr(module_self, LINGER_FUNCINT_BMM_COUNTER, iname_index+1)
    var_name = LINGER_MIX_INT8_MANUAL_ROUND_LAYERS + \
        '_function_bmm_'+str(iname_index)
    quant_layer = None
    bmm_output = None
    if hasattr(module_self, var_name):
        quant_layer = getattr(module_self, var_name)
    else:
        quant_layer = BmmInt(data_bits=8, mode=quant_mode,)
        quant_layer.training = module_self.training
        quant_layer = quant_layer.to(input.device)
        setattr(module_self, var_name, quant_layer)
    quant_layer.clamp_data = None
    quant_layer.o_bits = out_bits
    bmm_output = quant_layer(input, mat2)
    return bmm_output


class iqVar(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, dim, unbiased, keepdim, scale_x, local_scale_o, running_o, scale_o, training, quant_mode, prefix, dump, path):
        ctx.save_for_backward(x)
        ctx.value = dim, unbiased, keepdim
        x_int = x.quant_to_int8(scale_x())
        x_int = x_int.contiguous()
        scale_z_iq = local_scale_o
        momentum = 0.1
        bound_value = 127

        if training:
            running_o.mul_(1-momentum).add_(momentum * (bound_value/local_scale_o()))
            if quant_mode == QuantMode.QValue:
                scale_log = round(math.log(scale_z_iq(), 2))
                scale_z_iq = math.pow(2, scale_log)
            scale_z_iq = ScalerBuffer(scale_z_iq)
        else:
            assert running_o.data > 0, 'Must at least training one batch'
            if quant_mode == QuantMode.QValue:
                scale_log = round(math.log(bound_value/running_o.data, 2))
                scale_z_iq = math.pow(2, scale_log)
            else:
                scale_z_iq = np.float32(bound_value / running_o.data)
            scale_z_iq = ScalerBuffer(scale_z_iq)
            scale_o.fill_(scale_z_iq())
        y_int = None

        if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
            x_float = Quant.dequant(x_int, scale_x)
            y_float = torch_var(x_float, dim, unbiased, keepdim)
            y_int = (y_float * scale_z_iq()).round().int()
            y_int.clamp_(-128,127)

        y_float = Quant.dequant(y_int, scale_z_iq)

        if dump:
            name_list = ["input", "outputs", "q_input", "q_outputs"]
            attr_list = [x, y_float, x_int, y_int]
            Dump.dump_file(prefix, ".iqVar.", zip(name_list, attr_list), path)
        
        return from_torch_tensor(y_float, scale_z_iq(), 8)

    @staticmethod
    def backward(ctx, s):
        x, = ctx.saved_tensors
        dim, unbiased, keepdim  = ctx.value
        x = x.detach().clone().requires_grad_(True)
        grad = None
        with torch.enable_grad():
            y = torch_var(x, dim, unbiased, keepdim)
            grad = torch.autograd.grad(y, x, s)
        return grad[0], None, None, None, None, None, None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, x, dim, unbiased, keepdim, scale_x, local_scale_o, running_o, scale_o, training, quant_mode, prefix, dump, path):
        param_dict = {"scale_x_f": scale_x(), "scale_o_f": scale_o()}
        input_list = [x, ]
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict["platform_quant_s"] = platform_quant
        param_dict["castor_mode_s"] = "luna"
        param_dict["dim_i"] = dim
        param_dict["unbiased_i"] = unbiased
        op = None
        op = g.op("thinker::iqVar", *input_list, **param_dict)
        return op

class iqVarLayer(torch.nn.Module):
    def __init__(self, data_bits=8, mode=QuantMode.QValue,):
        super(iqVarLayer, self).__init__()
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.data_bits = data_bits
        self.mode = mode
        self.register_buffer('scale_o', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))

    def forward(self, input, dim, unbiased, keepdim, local_scale_o, quant_mode=QuantMode.QValue):
        scale_x = ScalerBuffer(input.scale_data)
        local_scale_o = ScalerBuffer(local_scale_o)
        scale_o = ScalerBuffer(self.scale_o)
        running_o = ScalerBuffer(self.running_o)
        if isinstance(input, IQTensor):
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits, self.mode)
            scale_x = ScalerBuffer(input.scale_data)
        z = iqVar.apply(input, dim, unbiased, keepdim, scale_x, local_scale_o, running_o, scale_o,
                            self.training, quant_mode, self.prefix, self.dump, self.path)
        self.scale_o.fill_(scale_o.data)
        self.running_o.fill_(running_o.data)
        return z

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)

def var(tensor, dim, unbiased=True, keepdim=False, *, out=None):
    is_iq_var = True
    if not isinstance(tensor,IQTensor):
        is_iq_var = False
    module_self = get_current_module()
    if module_self is None:
        is_iq_var = False   
    if not is_iq_var:
        return torch_var(tensor)

    assert tensor.bits == 8, 'iqvar only support 8bit'
    quant_mode = getattr(module_self, LINGER_MODE, QuantMode.QValue)
    iname_index = getattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER)
    setattr(module_self, LINGER_IQTENSOR_LAYER_COUNTER, iname_index+1)
    var_name = LINGER_MIX_INT8_MANUAL_ROUND_LAYERS + '_iqvar_' + str(iname_index)
    
    iq_layer = None
    if hasattr(module_self,var_name):
        iq_layer = getattr(module_self,var_name)
    else:
        iq_layer = iqVarLayer()
        iq_layer.training = module_self.training
        iq_layer = iq_layer.to(tensor.device)
        setattr(module_self,var_name,iq_layer)
    scale_z = None
    with torch.no_grad():
        z_f = torch_var(tensor, dim, unbiased, keepdim)
        max_z = torch.max(torch.abs(z_f))
        if max_z == 0:
            scale_z = 1.0
        else:
            scale_z = 127 / max_z.item()
    return iq_layer(tensor, dim, unbiased, keepdim, scale_z, quant_mode)


torch.max_pool2d = max_pool2d
torch.relu = relu
torch.max = iqmax

torch.relu_ = relu_
torch.transpose = transpose
torch.nn.functional.pad = pad

__all__ = ['torch_relu', 'torch_relu_', 'torch_max_pool2d', 'torch_cat', 'iqCatLayer', 'cat', 'torch_sigmoid', 'torch_sigmoid_',
           'iqSigmoidLayer', 'sigmoid', 'sigmoid_', 'torch_tanh', 'torch_tanh_', 'iqTanhLayer', 'tanh', 'tanh_', 'torch_clamp', 'torch_clamp_', 'iqClampLayer', 'clamp', 'clamp_',
           'dropout', 'pack_padded_sequence', 'pad_packed_sequence', 'torch_pack_padded_sequence', 'torch_pad_packed_sequence', 'bmm',
           'torch_softmax', 'softmaxIntLayer', 'softmax',
           'torch_logsoftmax', 'logsoftmaxIntLayer', 'logsoftmax', 'var', 'channel_shuffle_quant', 'channel_shuffle']



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from linger.utils import PlatFormQuant
import torch
import torch.nn as nn
import linger
import numpy as np
import os
if not os.path.exists('data.ignore'):
    os.mkdir('data.ignore')
def test_group_fuse_bn1d():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.conv = nn.Conv1d(10, 10, kernel_size=3, stride=1,
                         padding=1, bias=True, groups=2)
            self.bn = nn.BatchNorm1d(10)
            self.fc = nn.Linear(10*50, 100)
    
        def forward(self, x):
            x = self.conv(x)
            x = self.bn(x)
            n, c, l = x.shape
            x = x.view(n, c*l)
            x = self.fc(x)
            return x
    
    class Net1(nn.Module):
        def __init__(self):
            super(Net1, self).__init__()
            self.conv = nn.Conv1d(10, 10, kernel_size=3, stride=1,
                         padding=1, bias=True, groups=2)
            self.bn = nn.BatchNorm1d(10)
            self.fc = nn.Linear(10*50, 100)
    
        def forward(self, x):
            x = self.conv(x)
            x = self.bn(x)
            n, c, l = x.shape
            x = x.view(n, c*l)
            x = self.fc(x)
            return x
    
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)
    torch.cuda.set_device(0)
    aa = torch.randn(10, 10, 50).cuda()
    
    replace_tuple=(nn.Conv1d, nn.ConvTranspose2d, nn.Linear)
    net1 = Net().cuda()
    net1.eval()
    print(net1)
    net1(aa)
    torch.save(net1.state_dict(), 'data.ignore/model1d.pt.ignore')
    net1.train()
    out1 = net1(aa)
    
    net2 = Net1().cuda()
    
    linger.trace_layers(net2,net2, aa)
    linger.SetPlatFormQuant(platform_quant=PlatFormQuant.luna_quant)
    net3 = linger.init(net2, quant_modules=replace_tuple)
    #net3 = net2
    net3.load_state_dict(torch.load('data.ignore/model1d.pt.ignore'))
    out3 = net3(aa)
    
    assert out1.sum() - out3.sum() < 0.01
    
    


from io import BytesIO

import torch
import torch.nn
import torch.onnx

import onnx

from .scope import build_global_scope, build_onnx_scope_info
from .update_dequant import parser_dequant

torch_onnx_export = torch.onnx.export


def export(model, args, f, export_params=True, verbose=False, training=False,
           input_names=None, output_names=None, aten=False, export_raw_ir=False,
           operator_export_type=None, opset_version=12, _retain_param_name=True,
           do_constant_folding=True, example_outputs=None, strip_doc_string=True,
           dynamic_axes=None, keep_initializers_as_inputs=None, custom_opsets=None,
           enable_onnx_checker=True, use_external_data_format=False, is_update_dequant=True,
           is_input_quant=False, is_scoped_info=False, debug_dump=False):
    print('change onnx export to linger export')
    if is_scoped_info:
        scopes_info = {}
        scopes_info = build_global_scope(model)

    if is_update_dequant:
        if isinstance(args, tuple):
            model.eval()
            model(*args)
            args = list(args)
            args = tuple([arg if not isinstance(arg, float)
                         else torch.tensor(arg) for arg in args])
        else:
            model.eval()
            model(args)
        tmp = BytesIO()
        torch_onnx_export(model, args, tmp, export_params, verbose, training, input_names, output_names, aten, export_raw_ir,
                          operator_export_type, opset_version, _retain_param_name, do_constant_folding, example_outputs, strip_doc_string,
                          dynamic_axes, keep_initializers_as_inputs, custom_opsets, enable_onnx_checker, use_external_data_format)
        tmp.seek(0)
        onnx_model = onnx.load(tmp)

        if debug_dump:
            onnx.save(onnx_model, "debug_onnx_torch_export.onnx")

        if is_scoped_info:
            onnx_model = build_onnx_scope_info(onnx_model)
            if debug_dump:
                onnx.save(onnx_model, "debug_onnx_scoped_info.onnx")
        
        onnx_model = parser_dequant(onnx_model, is_input_quant)
        if debug_dump:
            onnx.save(onnx_model, "debug_onnx_update_dequant.onnx")

        return onnx.save(onnx_model, f)
    else:
        print("Error:is_update_dequant is not support now")

    if is_scoped_info:
        for _, enter_hook, leave_hook in scopes_info.values():
            enter_hook.remove()
            leave_hook.remove()
        scopes_info.clear()


__all__ = ['export']



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import linger
import numpy
import torch
import torch.nn as nn


def test_NormalizeLSTM_onnx_export():

    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()

            self.lstm = linger.NormalizeFastLSTM(
                100, 50, num_layers=1, batch_first=True, bidirectional=True)
            # self.lstm = nn.LSTM(
            #     100, 50, num_layers=1, batch_first=True, bidirectional=True)

        def forward(self, input, batch_lengths=None, initial_state=None):
            # input  (b t d)
            # x = nn.utils.rnn.pack_padded_sequence(x, batch_lengths, batch_first=True, enforce_sorted=False)

            # normalize
            x = (input, batch_lengths, True, False)
            x, hc = self.lstm(x, initial_state)
            x, _ = x

            # torch
            # x, hc = self.lstm(input, initial_state)

            # x, _ = nn.utils.rnn.pad_packed_sequence(x, batch_first=True)

            return x
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    numpy.random.seed(1)
    aa = torch.randn(10, 10, 100).cuda()
    label = torch.randint(10, (10, 10)).cuda()  # class=10
    mask = torch.ones(10, 10)
    for i in range(9):
        index = numpy.random.randint(5, 10)
        mask[i, index:] = 0
        label[i, index:] = -1

    input_lengths = mask.long().sum(1).cpu().numpy()
    input_lengths = torch.tensor(input_lengths)  # .cuda()

    batch_size = 10
    hidden_size = 50
    size = 2
    initial_state = (torch.zeros(size, batch_size, hidden_size).cuda(),
                     torch.zeros(size, batch_size, hidden_size).cuda())
    net = Net().cuda()

    net.eval()

    out1 = net(aa, input_lengths, initial_state)

    with torch.no_grad():
        net.eval()
        torch.onnx.export(net, (aa, input_lengths, initial_state), "data.ignore/normalize_torch_lstm.onnx", export_params=True,
                          opset_version=12, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)



import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..config import config
from ..quant import (Quant, normalize_bias_with_config,
                     normalize_data_with_config, normalize_weight_with_config)
from ..utils import Dump, PlatFormQuant, QuantMode, ScalerBuffer
from .iqtensor import (IQTensor, from_torch_tensor, platform_to_string,
                       quantlinear)
from .ops import ModuleIntConfig
from .requant import Requant


class LinearFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, weight, bias, data_bits, parameter_bits, running_x, running_w, running_o, eval_scale_x, eval_scale_w, eval_scale_o,
                momentum, training, prefix, dump, path, mode, o_bits, quant, is_not_from_iqtensor, ahead_relu, clamp_data, clamp_weight, clamp_bias, ahead_sigmoid):
        scale_o = None

        # venus limits
        n = input.shape[-1]
        if (data_bits == 8):
            assert math.ceil(4/4) * 4 * math.ceil(n/8) * 8 * data_bits/8 <= 64 * \
                1024, f"in LinearInt op, input shape must satisfy math.ceil(4/4) * 4 * math.ceil(n/8) * 8 * data_bits/8 <= 64*1024, but you have math.ceil({4}/4) * 4 * math.ceil({n}/8) * 8 * {data_bits/8} <= 64*1024"
        elif (data_bits == 16):
            assert math.ceil(4/4) * 4 * math.ceil(n/2) * 2 * data_bits/8 <= 64 * \
                1024, f"in LinearInt op, input shape must satisfy math.ceil(4/4) * 4 * math.ceil(n/2) * 2 * data_bits/8 <= 64*1024, but you have math.ceil({4}/4) * 4 * math.ceil({n}/2) * 2 * {data_bits/8} <= 64*1024"
        elif (data_bits == 32):
            assert math.ceil(2/2) * 2 * math.ceil(n/2) * 2 * data_bits/8 <= 64 * \
                1024, f"in LinearInt op, input shape must satisfy math.ceil(2/2) * 2 * math.ceil(n/2) * 2 * data_bits/8 <= 64*1024, but you have math.ceil({2}/2) * 2 * math.ceil({n}/2) * 2 * {data_bits/8} <= 64*1024"

        if training:
            ctx.clamp_data = clamp_data
            ctx.clamp_weight = clamp_weight
            ctx.clamp_bias = clamp_bias
            zero_point = input.zero_point if isinstance(input, IQTensor) else 0
            is_iq_tensor = True if isinstance(input, IQTensor) else False
            ctx.value = zero_point, is_iq_tensor
            ctx.bits = data_bits, parameter_bits, o_bits
            saved_tensors = [input, weight, bias]
            if isinstance(input, IQTensor):
                q_input, _, max_value_x = quant.quant(
                    input.data, data_bits, eval_scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
                scale_x = eval_scale_x
            else:
                q_input, scale_x, max_value_x = quant.quant(
                    input, data_bits, mode=mode, quant_data='input')
                running_x.mul_(1-momentum).add_(momentum*max_value_x)

            q_weight, scale_w, max_value_w = quant.quant(
                weight, parameter_bits, mode=mode, quant_data='weight')
            running_w.mul_(1-momentum).add_(momentum*max_value_w)
            q_input = q_input.float() if data_bits + parameter_bits <= 16 else q_input.double()
            q_weight = q_weight.float() if data_bits + \
                parameter_bits <= 16 else q_weight.double()
            q_outputs = F.linear(q_input, q_weight)
            outputs = None
            q_bias = None
            if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                assert mode == QuantMode.QValue, 'castor quant only support QValue and o_bits=None'
                if bias is not None:
                    q_bias = (bias * scale_w * scale_x + 0.5).floor()
                    if data_bits + parameter_bits <= 16:
                        q_bias = q_bias.float()
                    else:
                        q_bias = q_bias.double()
                    q_outputs += q_bias.reshape(-1)
                outputs = quant.dequant(q_outputs, scale_x*scale_w)
            else:
                assert False, "linger only support luna quant."

            # f_input = quant.dequant(q_input, scale_x)
            # f_weight = quant.dequant(q_weight, scale_w)
            # f_bias = None if bias is None else quant.dequant(q_bias, scale_x*scale_w)
            # saved_tensors = [f_input, f_weight, f_bias]

            out_tensor = outputs
            # ctx.save_for_backward(*saved_tensors)
            if o_bits is not None:
                outputs = normalize_data_with_config(outputs, clamp_data)
                out_tensor = outputs
                q_outputs, scale_o, max_value_o = quant.quant(
                    outputs, o_bits, mode=mode, quant_data='output', ahead_relu=ahead_relu)
                running_o.mul_(1-momentum).add_(momentum*max_value_o)
                outputs = quant.dequant(q_outputs, scale_o)
            ctx.scale = scale_x, scale_w, scale_o
            saved_tensors += [out_tensor]
            ctx.save_for_backward(*saved_tensors)

        else:
            assert running_x > 0, 'invalid running_x, please finetune training'
            q_weight = None
            scale_o = eval_scale_o
            if weight.dtype == torch.float32:
                scale_x = ScalerBuffer(quant.running_to_scale(
                    running_x, data_bits, mode=mode))
                q_weight, scale_w, _ = quant.quant(
                    weight, parameter_bits, mode=mode, quant_data='weight')
                scale_w = ScalerBuffer(scale_w)
                if o_bits is not None:
                    assert running_o > 0, 'invalid running_o <= 0, please finetune training'
                    scale_o = ScalerBuffer(quant.running_to_scale(
                        running_o, o_bits, mode=mode))
            else:
                scale_x = eval_scale_x
                scale_w = eval_scale_w
                q_weight = weight.double()
                if o_bits is not None:
                    scale_o = eval_scale_o

            if isinstance(input, IQTensor):
                scale_x = eval_scale_x
                q_input, _, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
            else:
                q_input, scale_x, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=mode, quant_data='input')
            q_input = q_input.double()
            q_weight = q_weight.double()
            q_outputs = F.linear(q_input, q_weight)

            outputs = None
            q_bias = None
            if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                assert mode == QuantMode.QValue, 'castor quant only support QValue and o_bits=None'
                if bias is not None:
                    if bias.dtype == torch.float32:
                        q_bias = (bias * scale_w * scale_x + 0.5).floor()
                        if data_bits + parameter_bits <= 16:
                            q_bias = q_bias.float().double()
                        else:
                            q_bias = q_bias.double()
                    else:
                        q_bias = bias.double()
                    q_outputs += q_bias.reshape(-1)
                outputs = quant.dequant(q_outputs, scale_x*scale_w)
            else:
                assert False, "linger only support luna quant."
            if o_bits is not None:
                if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                    q_outputs, _, _ = quant.quant(
                        outputs, o_bits, scale_o, mode=mode, quant_data='output')
                outputs = quant.dequant(q_outputs, scale_o)
            if dump:
                if bias is not None:
                    name_list = ["input", "weight", "bias", "outputs", "q_input", "q_weight", "q_bias",
                                 "q_outputs", "running_x", "running_w", "running_o", "scale_x", "scale_w", "scale_o"]
                    attr_list = [input, weight, bias, outputs, q_input, q_weight, q_bias, q_outputs,
                                 running_x.data, running_w.data, running_o.data, scale_x.data, scale_w.data, scale_o.data]
                    Dump.dump_file(prefix, ".LinearInt.",
                                   zip(name_list, attr_list), path)
                else:
                    name_list = ["input", "weight", "outputs", "q_input", "q_weight", "q_outputs",
                                 "running_x", "running_w", "running_o", "scale_x", "scale_w", "scale_o"]
                    attr_list = [input, weight, outputs, q_input, q_weight, q_outputs, running_x.data,
                                 running_w.data, running_o.data, scale_x.data, scale_w.data, scale_o.data]
                    Dump.dump_file(prefix, ".LinearInt.",
                                   zip(name_list, attr_list), path)
        if o_bits is None:
            return outputs
        elif isinstance(scale_o, float):
            return from_torch_tensor(outputs, scale_o, o_bits)
        elif isinstance(scale_o, torch.Tensor):
            return from_torch_tensor(outputs, scale_o.item(), o_bits)
        else:
            return from_torch_tensor(outputs, scale_o.data, o_bits)

    @staticmethod
    def backward(ctx, gradOutput):
        clamp_data = ctx.clamp_data
        data_bits, parameter_bits, o_bits = ctx.bits
        zero_point, is_iq_tensor = ctx.value
        scale_x, scale_w, scale_o = ctx.scale
        input, weights, bias, outputs = ctx.saved_tensors
        if is_iq_tensor:
            f_input = input.data
        else:
            q_input, _, _ = Quant.quant(
                input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input')
            f_input = Quant.dequant(q_input, scale_x)

        f_input = f_input.detach().clone().requires_grad_(True)
        q_weights, _, _ = Quant.quant(
            weights.data, parameter_bits, scale_w, mode=QuantMode.QValue, quant_data='weight')
        f_weights = Quant.dequant(q_weights, scale_w)

        f_weights = f_weights.detach().clone().requires_grad_(True)
        bias = None if bias is None else bias.detach().clone().requires_grad_(True)
        gradInput = gradWeight = gradBias = None

        with torch.enable_grad():
            z = F.linear(f_input, f_weights, bias)
            if o_bits is not None:
                z = normalize_data_with_config(z, clamp_data)
            if bias is not None:
                gradInput, gradWeight, gradBias = torch.autograd.grad(
                    z, (f_input, f_weights, bias), gradOutput)
            else:
                gradInput, gradWeight, = torch.autograd.grad(
                    z, (f_input, f_weights), gradOutput)

        return gradInput, gradWeight, gradBias, None, None, None, None, None, None, None, None, None, None, None, None,\
            None, None, None, None, None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, input, weight, bias, data_bits, parameter_bits, running_x, running_w, running_o, scale_x, scale_w, scale_o, momentum, training,
                 prefix, dump, path, mode, o_bits, quant, is_not_from_iqtensor, ahead_relu, clamp_data, clamp_weight, clamp_bias, ahead_sigmoid):
        op_inner = None
        if is_not_from_iqtensor:
            platform_quant = platform_to_string(
                config.PlatFormQuant.platform_quant)
            op_inner = quantlinear(g, input, scale_x(),
                                   platform_quant, data_bits)
            input_list = [op_inner, weight]
        else:
            input_list = [input, weight]
        param_dict = {'scale_x_f': scale_x(), 'scale_w_f': scale_w(),
                      'data_bits_i': data_bits, 'parameter_bits_i': parameter_bits}
        if bias is not None:
            input_list.append(bias)
        if o_bits is not None:
            param_dict['scale_o_f'] = scale_o()
            param_dict['o_bits_i'] = o_bits

        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict['platform_quant_s'] = platform_quant

        return g.op("thinker::LinearInt", *input_list, **param_dict)


class LinearInt(nn.Linear, ModuleIntConfig):
    r"""实现LinearInt的量化训练与测试，继承自nn.Linear,

    Args:
        in_features out_features bias
        与nn.Linear一致的参数
        data_bits(int): 输入量化位数
        parameter_bits(int): 参数量化位数
        mode(Enum): 量化方式，支持MaxValue与Qvalue
        o_bits(int, default=None):输出量化位数
        clamp_data(float or None): 针对输出的clamp数值
        clamp_weight(float or None): 针对转为weight的clamp数值
        clamp_bias(float or None): 与clamp_weight一致
        ahead_relu(bool): 是否做融合relu之后的数值统计scale

    """

    def __init__(self, in_features, out_features, bias=True, data_bits=8, parameter_bits=8, mode=QuantMode.QValue,
                 o_bits=None, clamp_data=None, clamp_weight=None, clamp_bias=None, ahead_relu=False, ahead_sigmoid=False):
        # assert data_bits == parameter_bits, "data_bits and parameter_bits must be equal"
        assert data_bits in (8, 16, 32), "data_bits only support 8, 16, 32"
        nn.Linear.__init__(self, in_features, out_features, bias)
        ModuleIntConfig.__init__(
            self, data_bits=data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=o_bits)

        self.prefix = ""
        self.dump = False
        self.path = ""
        self.momentum = 0.1
        self.is_not_from_iqtensor = True
        self.clamp_data = clamp_data
        self.clamp_weight = clamp_weight
        self.clamp_bias = clamp_bias
        self.ahead_relu = ahead_relu
        self.ahead_sigmoid = ahead_sigmoid
        self.mode = mode
        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('running_w', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('scale_w', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))

    def forward(self, input):
        scale_x = ScalerBuffer(self.scale_x)
        running_x = ScalerBuffer(self.running_x)
        if isinstance(input, IQTensor):
            self.is_not_from_iqtensor = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits, self.mode)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)
        running_w = ScalerBuffer(self.running_w)
        running_o = ScalerBuffer(self.running_o)
        scale_w = ScalerBuffer(self.scale_w)
        scale_o = ScalerBuffer(self.scale_o)
        weight = self.weight
        bias = self.bias
        if weight.dtype == torch.float32:
            weight = normalize_weight_with_config(
                self.weight, self.clamp_weight, self.training)
            if self.bias is not None:
                bias = normalize_bias_with_config(
                    self.bias, self.clamp_bias, self.training)
        ret = LinearFunction.apply(input, weight, bias,
                                   self.data_bits, self.parameter_bits,
                                   running_x, running_w, running_o, scale_x, scale_w, scale_o,
                                   self.momentum, self.training, self.prefix, self.dump, self.path,
                                   self.quant_mode, self.o_bits, self.quant, self.is_not_from_iqtensor, self.ahead_relu,
                                   self.clamp_data, self.clamp_weight, self.clamp_bias, self.ahead_sigmoid)
        self.running_x.fill_(running_x())
        self.running_w.fill_(running_w())
        self.running_o.fill_(running_o())
        self.scale_x.fill_(scale_x())
        self.scale_w.fill_(scale_w())
        self.scale_o.fill_(scale_o())
        return ret

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        return ModuleIntConfig.state_dict_global(self, destination, prefix, keep_vars)

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)

    def extra_repr(self):
        s = nn.Linear.extra_repr(self)
        extra_s = ',clamp_data:{clamp_data},clamp_weight:{clamp_weight},clamp_bias:{clamp_bias},ahead_relu:{ahead_relu},ahead_sigmoid:{ahead_sigmoid}'.format(
            **self.__dict__)
        extra_s += ',data_bits:{data_bits},parameter_bits:{parameter_bits},o_bits:{o_bits}'.format(
            **self.__dict__)
        return s+extra_s



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..config import config
from ..quant import Quant
from ..utils import Dump, PlatFormQuant, QuantMode, ScalerBuffer
from .iqtensor import (IQTensor, from_torch_tensor, platform_to_string,
                       quantlinear)
from .ops import ModuleIntConfig
from .requant import Requant


class AvgPool2dIntFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, kernel_size, stride, padding, ceil_mode, count_include_pad, divisor_override,
                data_bits, training, momentum, running_x, running_o, eval_scale_x, eval_scale_o, prefix, dump, path, mode, o_bits, quant,
                is_not_from_iqtensor):
        # venus limits
        assert data_bits in (
            4, 8), f"in AvgPool2d op, AvgPool2d data_bits only support 4/8 bits, but you have data_bits {data_bits}"
        assert o_bits in (
            4, 8, 16), f"in AvgPool2d op, AvgPool2d o_bits only support 4/8/16 bits, but you have o_bits {o_bits}"

        scale_x = None
        scale_o = None
        if training:
            ctx.save_for_backward(input)
            ctx.params = [kernel_size, stride, padding,
                          ceil_mode, count_include_pad, divisor_override]
            zero_point = input.zero_point if isinstance(input, IQTensor) else 0
            is_iq_tensor = True if isinstance(input, IQTensor) else False
            # ctx.bits = data_bits
            if isinstance(input, IQTensor):
                q_input, _, max_value_x = quant.quant(
                    input.data, data_bits, eval_scale_x, mode=QuantMode.QValue, quant_data='input',
                    iq_zero_point=input.zero_point)
                scale_x = eval_scale_x
            else:
                q_input, scale_x, max_value_x = quant.quant(
                    input.data, data_bits, mode=mode, quant_data='input')
                scale_x = ScalerBuffer(scale_x)
                running_x.mul_(1-momentum).add_(momentum*max_value_x)
            q_input = q_input.float() if data_bits <= 8 else q_input.double()
            q_outputs_float = F.avg_pool2d(q_input.contiguous(), kernel_size, stride=stride, padding=padding, ceil_mode=ceil_mode,
                                           count_include_pad=count_include_pad, divisor_override=divisor_override)
            if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                if isinstance(kernel_size, tuple):
                    kernel_area = kernel_size[0] * kernel_size[1]
                else:
                    kernel_area = kernel_size * kernel_size
                q_outputs = (
                    (q_outputs_float * kernel_area).round() / kernel_area + 0.5).floor()
                bound_value = math.pow(2, data_bits-1) - 1
                q_outputs.clamp_(-bound_value-1, bound_value)
            else:
                assert False, "linger only support luna quant."
            outputs = quant.dequant(q_outputs, scale_x)
            if o_bits is not None:
                running_o.fill_(running_x())
                scale_o = scale_x
            ctx.value = scale_x, data_bits, zero_point, is_iq_tensor
        else:
            assert running_x > 0, 'invalid running_x <= 0, please fintune first'
            scale_x = None
            scale_o = None
            scale_x = quant.running_to_scale(running_x, data_bits, mode=mode)
            scale_x = ScalerBuffer(scale_x)
            if isinstance(input, IQTensor):
                scale_x = eval_scale_x
                q_input, _, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
            else:
                q_input, _, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=mode, quant_data='input')
            q_input = q_input.double()
            q_outputs = F.avg_pool2d(q_input.contiguous(), kernel_size, stride=stride, padding=padding, ceil_mode=ceil_mode,
                                     count_include_pad=count_include_pad, divisor_override=divisor_override)
            if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                if isinstance(kernel_size, tuple):
                    kernel_area = kernel_size[0] * kernel_size[1]
                else:
                    kernel_area = kernel_size * kernel_size
                q_outputs = ((q_outputs * kernel_area).round() /
                             kernel_area + 0.5).floor()
                bound_value = math.pow(2, data_bits-1) - 1
                q_outputs.clamp_(-bound_value-1, bound_value)
            else:
                assert False, "linger only support luna quant."
            outputs = quant.dequant(q_outputs, scale_x)
            scale_o = eval_scale_o
            if o_bits is not None:
                scale_o.fill_(scale_x())
            if dump:
                name_list = ["input", "outputs", "q_input",  "q_outputs",
                             "scale_x", "scale_o", "running_x", "running_o"]
                attr_list = [input, outputs, q_input, q_outputs,
                             scale_x.data, scale_o.data, running_x.data,  running_o.data]
                Dump.dump_file(prefix, ".AvgPool2dInt.",
                               zip(name_list, attr_list), path)

        if o_bits is None:
            return outputs
        elif isinstance(scale_o, float):
            return from_torch_tensor(outputs, scale_o, o_bits)
        elif isinstance(scale_o, torch.Tensor):
            return from_torch_tensor(outputs, scale_o.item(), o_bits)
        else:
            return from_torch_tensor(outputs, scale_o.data, o_bits)

    @staticmethod
    def backward(ctx, gradoutput):
        input, = ctx.saved_tensors
        scale_x, data_bits, zero_point, is_iq_tensor = ctx.value
        kernel_size, stride, padding, ceil_mode, count_include_pad, divisor_override = ctx.params
        if is_iq_tensor:
            backward_input = input.data
        else:
            q_input, _, _ = Quant.quant(
                input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input')
            backward_input = Quant.dequant(q_input, scale_x)
        backward_input = backward_input.detach().clone().requires_grad_(True)
        grad = None
        with torch.enable_grad():
            output = F.avg_pool2d(backward_input, kernel_size, stride,
                                  padding, ceil_mode, count_include_pad, divisor_override)
            grad = torch.autograd.grad(output, backward_input, gradoutput)
        return grad[0], None, None, None, None, None, None, \
            None, None, None, None, None, None, None, None, None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, input, kernel_size, stride, padding, ceil_mode, count_include_pad, divisor_override,
                 data_bits, training, momentum, running_x, running_o, scale_x, scale_o, prefix, dump, path, mode, o_bits, quant, is_not_from_iqtensor):
        op_inner = None
        if is_not_from_iqtensor:
            platform_quant = platform_to_string(
                config.PlatFormQuant.platform_quant)
            op_inner = quantlinear(g, input, scale_x(),
                                   platform_quant, data_bits)
        if isinstance(padding, int):
            padding = [padding]*2
        if isinstance(kernel_size, int):
            kernel_size = [kernel_size]*2
        if isinstance(stride, int):
            stride = [stride]*2
        paddings = padding + padding
        param_dict = {'kernel_shape_i': kernel_size, 'strides_i': stride, 'pads_i': paddings, 'ceil_mode_i': ceil_mode, 'data_bits_i': data_bits,
                      'scale_x_f': scale_x()}
        if is_not_from_iqtensor:
            input_list = [op_inner, ]
        else:
            input_list = [input, ]
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict['platform_quant_s'] = platform_quant
        if o_bits is not None:
            param_dict['scale_o_f'] = scale_o()
            param_dict['o_bits_i'] = o_bits
        return g.op("thinker::AvgPool2dInt", *input_list, **param_dict)


class AvgPool2dInt(nn.AvgPool2d, ModuleIntConfig):
    r"""实现AvgPool2dInt的量化训练与测试，继承自nn.AvgPool2d,
    Args: 
        kernel_size stride padding, ceil_mode count_include_pad divisor_override
        与nn.AvgPool2d一致
        data_bits(int, default=8): 输入的量化位数
        mode(Enum): 量化方式，支持MaxValue与Qvalue
        o_bits(int, default=None):输出量化位数
    Notes:
        支持luna_quant
        luna_quant: x = (x + 0.5).floor()

    """

    def __init__(self, kernel_size, stride, padding=0, ceil_mode=False,
                 count_include_pad=True, divisor_override=None,
                 data_bits=8, mode=QuantMode.QValue, o_bits=None):
        nn.AvgPool2d.__init__(self, kernel_size, stride, padding,
                              ceil_mode, count_include_pad, divisor_override)
        ModuleIntConfig.__init__(
            self, data_bits=data_bits, mode=mode, o_bits=o_bits)
        self.momentum = 0.1
        self.is_not_from_iqtensor = True
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))

    def forward(self, input):
        scale_x = ScalerBuffer(self.scale_x)
        running_x = ScalerBuffer(self.running_x)
        assert (self.o_bits is None or self.o_bits ==
                self.data_bits), 'AvgPool2dInt out_bits must equal data_bits'
        if isinstance(input, IQTensor):
            self.is_not_from_iqtensor = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)
        running_o = ScalerBuffer(self.running_o)
        scale_o = ScalerBuffer(self.scale_o)
        output = AvgPool2dIntFunction.apply(input.contiguous(), self.kernel_size,
                                            self.stride, self.padding, self.ceil_mode,
                                            self.count_include_pad, self.divisor_override,
                                            self.data_bits, self.training, self.momentum,
                                            running_x, running_o, scale_x, scale_o, self.prefix, self.dump, self.path,
                                            self.quant_mode, self.o_bits, self.quant, self.is_not_from_iqtensor)
        self.running_x.fill_(running_x())
        self.running_o.fill_(running_o())
        self.scale_x.fill_(scale_x())
        self.scale_o.fill_(scale_o())
        return output

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        return ModuleIntConfig.state_dict_global(self, destination, prefix, keep_vars)

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)



import numpy as np
import torch
import torch.nn as nn

import linger


def test_var():
    class Model(nn.Module):
        def __init__(self):
            super(Model, self).__init__()
            self.conv = nn.Conv2d(10, 10, kernel_size=3, stride=1,
                                  padding=1, bias=True)
            self.fc = nn.Linear(1, 100)

        def forward(self, x):
            x = self.conv(x)
            x = torch.var(x, 1, False).reshape(-1)
            x = x.unsqueeze(-1)
            x = self.fc(x)
            return x

    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    np.random.seed(1)

    dummy_input = torch.randn(1, 10, 10, 10).cuda()
    target = torch.ones(1, 100).cuda()
    criterion = nn.MSELoss()

    replace_tuple = (nn.Conv2d, nn.Linear, nn.AvgPool2d)
    model = Model().cuda()

    linger.SetPlatFormQuant(platform_quant=linger.PlatFormQuant.luna_quant)
    linger.SetIQTensorVar(True)
    model = linger.init(model, quant_modules=replace_tuple,
                        mode=linger.QuantMode.QValue)

    optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
    loss = None
    for i in range(20):
        optimizer.zero_grad()
        out = model(dummy_input)
        print(out)
        loss = criterion(out, target)
        if i % 1 == 0:
            print('loss: ', loss)
        loss.backward()
        optimizer.step()

    with torch.no_grad():
        torch.onnx.export(model, dummy_input, "./data.ignore/var.onnx", export_params=True, opset_version=11,
                          operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)



import os
import sys

cur_dir = os.path.abspath(os.path.dirname(__file__))
sys.path.append(os.path.join(cur_dir, "lib"))

from .__version import __version__, version_info
from .config import *
from .conv_bn_fuser import EmptyBatchNorm, FuseBNIntoConv, FuseConvBNAheadRelu
from .dumper import Dumper
from .initialize import (DefaultQuantIntXOP, disable_quant, init, quant_module,
                         quant_module_by_type, quant_tensor)
from .layer_normalizer import (disable_normalize, normalize_layers,
                               normalize_module)
from .layer_tracer import trace_layers
from .modules import (NormalizeBatchNorm2d, NormalizeConv1d, NormalizeConv2d,
                      NormalizeConvBN1d, NormalizeConvBN2d,
                      NormalizeConvTranspose2d, NormalizeEmbedding,
                      NormalizeFastGRU, NormalizeFastLSTM, NormalizeLayerNorm,
                      NormalizeLinear)
from .onnx import parser_dequant
from .ops import *
from .quant import *
from .tools import fix_dequant, onnx_quant, wb_analyse
from .utils import *

name = "linger"



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import torch
import torch.nn as nn
from torch import _VF
from torch.nn.utils.rnn import PackedSequence
from torch.onnx import is_in_onnx_export

from ..quant import NormalizeFunction


class LSTMOnnxFakeFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, lengths, hidden_state, cell_state, weight_ih, weight_hh, bias_ih, bias_hh,
                weight_ih_reverse, weight_hh_reverse, bias_ih_reverse, bias_hh_reverse,
                input_size, hidden_size, num_layers, batch_first, dropout, bidirectional, bias_B, bias_B_reverse,
                hidden_state_forward, hidden_state_reverse, cell_state_forward, cell_state_reverse,
                weight_ih_bi, weight_hh_bi, bias_B_bi):
        output = None
        hidden_state = None
        cell_state = None
        batch_size = None
        seq_length = None
        num_directions = 2 if bidirectional else 1
        if batch_first:
            batch_size = input.size(0)
            seq_length = input.size(
                1) if lengths is None else torch.max(lengths)
            output = torch.randn(batch_size, seq_length,
                                 hidden_size*num_directions, device=input.device)
        else:
            batch_size = input.size(1)
            seq_length = input.size(
                0) if lengths is None else torch.max(lengths)
            output = torch.randn(seq_length, batch_size,
                                 hidden_size*num_directions, device=input.device)
        hidden_state = torch.zeros(
            num_directions, batch_size, hidden_size, device=input.device)
        cell_state = torch.zeros(
            num_directions, batch_size, hidden_size, device=input.device)
        return output, hidden_state, cell_state

    @staticmethod
    def backward(ctx, gradOutput, gradHidden, gradCell):

        return None, None, None, None, None, None, None,\
            None, None, None, None,\
            None, None, None, None, None, None,\
            None, None, None,\
            None, None, None, None, None, None, None,\
            None, None, None, None, None, None, None,\
            None, None, None, None, None, None, None,\
            None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, input, lengths, hidden_state, cell_state, weight_ih, weight_hh, bias_ih, bias_hh,
                 weight_ih_reverse, weight_hh_reverse, bias_ih_reverse, bias_hh_reverse,
                 input_size, hidden_size, num_layers, batch_first, dropout, bidirectional, bias_B, bias_B_reverse,
                 hidden_state_forward, hidden_state_reverse, cell_state_forward, cell_state_reverse,
                 weight_ih_bi, weight_hh_bi, bias_B_bi):
        if num_layers != 1:
            assert False, "Current intx not support num_layer!=1 onnx export !"
        param_dict = {'hidden_size_i': hidden_size, 'direction_s': "forward"}
        param_bidirectional_dict = {}
        if bidirectional:
            param_bidirectional_dict = {
                'hidden_size_i': hidden_size, 'direction_s': "bidirectional"}

        if batch_first:
            input = g.op("Transpose", *[input], perm_i=(1, 0, 2))

        input_list = [input, weight_ih, weight_hh]
        input_bidirectional_list = [input, weight_ih_bi, weight_hh_bi]
        if bias_ih is not None and bias_hh is not None:
            input_list.append(bias_B)
            input_bidirectional_list.append(bias_B_bi)

        param_dict['outputs'] = 3
        param_bidirectional_dict['outputs'] = 3

        if lengths is None and hidden_state is None:
            if not bidirectional:
                lstm, hidden, cell = g.op("LSTM", *input_list, **param_dict)
                lstm = g.op("Squeeze", *[lstm], axes_i=(1,))
                if batch_first:
                    lstm = g.op("Transpose", *[lstm], perm_i=(1, 0, 2))
            else:
                lstm_bi, hidden, cell = g.op(
                    "LSTM", *input_bidirectional_list, **param_bidirectional_dict)
                if batch_first:
                    lstm_bi = g.op(
                        "Transpose", *[lstm_bi], perm_i=(0, 2, 1, 3))
                args = [lstm_bi]
                lstm_bi = g.op('Reshape', *args, g.op('Constant',
                               value_t=torch.LongTensor([0, 0, -1])))
                lstm = g.op("Transpose", *[lstm_bi], perm_i=(1, 0, 2))

        elif lengths is not None and hidden_state is None:
            input_list.append(lengths)
            input_bidirectional_list.append(lengths)
            if not bidirectional:
                lstm, hidden, cell = g.op("LSTM", *input_list, **param_dict)
                lstm = g.op("Squeeze", *[lstm], axes_i=(1,))
                if batch_first:
                    lstm = g.op("Transpose", *[lstm], perm_i=(1, 0, 2))
            else:
                lstm_bi, hidden, cell = g.op(
                    "LSTM", *input_bidirectional_list, **param_bidirectional_dict)
                if batch_first:
                    lstm_bi = g.op(
                        "Transpose", *[lstm_bi], perm_i=(0, 2, 1, 3))
                args = [lstm_bi]
                lstm_bi = g.op('Reshape', *args, g.op('Constant',
                               value_t=torch.LongTensor([0, 0, -1])))
                lstm = g.op("Transpose", *[lstm_bi], perm_i=(1, 0, 2))

        else:
            input_list.append(lengths)
            input_list.append(hidden_state)
            input_list.append(cell_state)
            input_bidirectional_list.append(lengths)
            input_bidirectional_list.append(hidden_state)
            input_bidirectional_list.append(cell_state)
            if not bidirectional:
                lstm, hidden, cell = g.op("LSTM", *input_list, **param_dict)
                lstm = g.op("Squeeze", *[lstm], axes_i=(1,))
                if batch_first:
                    lstm = g.op("Transpose", *[lstm], perm_i=(1, 0, 2))
            else:
                lstm_bi, hidden, cell = g.op(
                    "LSTM", *input_bidirectional_list, **param_bidirectional_dict)
                if batch_first:
                    lstm_bi = g.op(
                        "Transpose", *[lstm_bi], perm_i=(0, 2, 1, 3))
                args = [lstm_bi]
                lstm_bi = g.op('Reshape', *args, g.op('Constant',
                               value_t=torch.LongTensor([0, 0, -1])))
                lstm = g.op("Transpose", *[lstm_bi], perm_i=(1, 0, 2))

        return lstm, hidden, cell


class NormalizeFastLSTM(nn.LSTM):

    def __init__(self, input_size, hidden_size, num_layers=1, bias=True, batch_first=False, dropout=0, bidirectional=False,
                 normalize_data=None, normalize_weight=None, normalize_bias=None):
        nn.LSTM.__init__(self, input_size, hidden_size, num_layers,
                         bias, batch_first, dropout, bidirectional)
        self.normalize_data = normalize_data
        self.normalize_weight = normalize_weight
        self.normalize_bias = normalize_bias

    def forward(self, input, hx=None):
        if not is_in_onnx_export():
            if self.num_layers != 1:
                assert False, "Intx-NormalizeLSTM don't support num_layer!=1 !"
            orig_input = input
            if isinstance(orig_input, PackedSequence):
                input, batch_sizes, sorted_indices, unsorted_indices = input
                max_batch_size = batch_sizes[0]
                max_batch_size = int(max_batch_size)
            elif isinstance(orig_input, tuple):
                input, lengths, batch_first, enforce_sorted = orig_input
                packed_input = torch.nn.utils.rnn.pack_padded_sequence(
                    input, lengths, batch_first, enforce_sorted)
                input, batch_sizes, sorted_indices, unsorted_indices = packed_input
                max_batch_size = batch_sizes[0]
                max_batch_size = int(max_batch_size)
            else:
                batch_sizes = None
                max_batch_size = input.size(
                    0) if self.batch_first else input.size(1)
                sorted_indices = None
                unsorted_indices = None

            if hx is None:
                num_directions = 2 if self.bidirectional else 1
                zeros = torch.zeros(self.num_layers * num_directions,
                                    max_batch_size, self.hidden_size,
                                    dtype=input.dtype, device=input.device)
                hx = (zeros, zeros)
            else:
                hx = self.permute_hidden(hx, sorted_indices)

            flat_weights = []
            direct = 0
            self.check_forward_args(input, hx, batch_sizes)
            weight_ih = self.weight_ih_l0_reverse if direct == 1 else self.weight_ih_l0
            weight_hh = self.weight_hh_l0_reverse if direct == 1 else self.weight_hh_l0
            if self.bias:
                bias_ih = self.bias_ih_l0_reverse if direct == 1 else self.bias_ih_l0
                bias_hh = self.bias_hh_l0_reverse if direct == 1 else self.bias_hh_l0

            if self.normalize_weight is not None:
                weight_ih = NormalizeFunction.apply(
                    weight_ih, self.normalize_weight, self.training)
                weight_hh = NormalizeFunction.apply(
                    weight_hh, self.normalize_weight, self.training)
            if bias_ih is not None and self.normalize_bias is not None:
                bias_ih = NormalizeFunction.apply(
                    bias_ih, self.normalize_bias, self.training)
                bias_hh = NormalizeFunction.apply(
                    bias_hh, self.normalize_bias, self.training)

            flat_weights.extend([weight_ih, weight_hh])
            if self.bias:
                flat_weights.extend([bias_ih, bias_hh])

            if self.bidirectional:
                direct = 1
                self.check_forward_args(input, hx, batch_sizes)
                weight_ih = self.weight_ih_l0_reverse if direct == 1 else self.weight_ih_l0
                weight_hh = self.weight_hh_l0_reverse if direct == 1 else self.weight_hh_l0
                if self.bias:
                    bias_ih = self.bias_ih_l0_reverse if direct == 1 else self.bias_ih_l0
                    bias_hh = self.bias_hh_l0_reverse if direct == 1 else self.bias_hh_l0
                if self.normalize_weight is not None:
                    weight_ih = NormalizeFunction.apply(
                        weight_ih, self.normalize_weight, self.training)
                    weight_hh = NormalizeFunction.apply(
                        weight_hh, self.normalize_weight, self.training)
                if bias_ih is not None and self.normalize_bias is not None:
                    bias_ih = NormalizeFunction.apply(
                        bias_ih, self.normalize_bias, self.training)
                    bias_hh = NormalizeFunction.apply(
                        bias_hh, self.normalize_bias, self.training)
                flat_weights.extend([weight_ih, weight_hh])
                if self.bias:
                    flat_weights.extend([bias_ih, bias_hh])

            if batch_sizes is None:
                result = _VF.lstm(input, hx,  flat_weights, self.bias, self.num_layers,
                                  self.dropout, self.training, self.bidirectional, self.batch_first)
            else:
                result = _VF.lstm(input, batch_sizes, hx,  flat_weights, self.bias,
                                  self.num_layers, self.dropout, self.training, self.bidirectional)
            output = result[0]
            hidden = result[1:]
            if self.normalize_data is not None:
                output = NormalizeFunction.apply(
                    output, self.normalize_data, self.training, True)

            if isinstance(orig_input, PackedSequence):
                output_packed = PackedSequence(
                    output, batch_sizes, sorted_indices, unsorted_indices)
                return output_packed, self.permute_hidden(hidden, unsorted_indices)
            elif isinstance(orig_input, tuple):
                output_packed = PackedSequence(
                    output, batch_sizes, sorted_indices, unsorted_indices)
                output, lengths = torch.nn.utils.rnn.pad_packed_sequence(
                    output_packed, self.batch_first)
                return (output, lengths), self.permute_hidden(hidden, unsorted_indices)
            else:
                return output, self.permute_hidden(hidden, unsorted_indices)
        else:
            orig_input = input
            if isinstance(orig_input, PackedSequence):
                assert False, "LSTM don't support PackedSequence as input for onnx export!"

            if isinstance(orig_input, tuple):
                input, lengths, _, _ = orig_input
            else:
                input = orig_input
                lengths = None

            bias_ih = None
            bias_hh = None
            bias_ih_reverse = None
            bias_hh_reverse = None
            weight_ih = self.weight_ih_l0
            weight_hh = self.weight_hh_l0
            weight_ih_reverse = weight_ih
            weight_hh_reverse = weight_hh

            if self.bias:
                bias_ih = self.bias_ih_l0
                bias_hh = self.bias_hh_l0
                bias_ih_reverse = bias_ih
                bias_hh_reverse = bias_hh
            if self.bidirectional:
                weight_ih_reverse = self.weight_ih_l0_reverse
                weight_hh_reverse = self.weight_hh_l0_reverse
                bias_ih_reverse = self.bias_ih_l0_reverse
                bias_hh_reverse = self.bias_hh_l0_reverse

            hidden_state = None
            cell_state = None
            if hx is not None:
                hidden_state, cell_state = hx
            output = None
            hy = None
            cy = None

            weight_ih_chunk = weight_ih.chunk(4, 0)
            weight_ih = torch.cat(
                [weight_ih_chunk[0], weight_ih_chunk[3], weight_ih_chunk[1], weight_ih_chunk[2]], dim=0)
            weight_ih = weight_ih.unsqueeze(0)
            weight_hh_chunk = weight_hh.chunk(4, 0)
            weight_hh = torch.cat(
                [weight_hh_chunk[0], weight_hh_chunk[3], weight_hh_chunk[1], weight_hh_chunk[2]], dim=0)
            weight_hh = weight_hh.unsqueeze(0)

            if self.normalize_weight is not None:
                weight_ih = NormalizeFunction.apply(
                    weight_ih, self.normalize_weight, self.training)
                weight_hh = NormalizeFunction.apply(
                    weight_hh, self.normalize_weight, self.training)

            weight_ih_bi = None
            weight_hh_bi = None

            if self.bidirectional:
                weight_ih_chunk_reverse = weight_ih_reverse.chunk(4, 0)
                weight_ih_reverse = torch.cat(
                    [weight_ih_chunk_reverse[0], weight_ih_chunk_reverse[3], weight_ih_chunk_reverse[1], weight_ih_chunk_reverse[2]], dim=0)
                weight_ih_reverse = weight_ih_reverse.unsqueeze(0)
                weight_hh_chunk_reverse = weight_hh_reverse.chunk(4, 0)
                weight_hh_reverse = torch.cat(
                    [weight_hh_chunk_reverse[0], weight_hh_chunk_reverse[3], weight_hh_chunk_reverse[1], weight_hh_chunk_reverse[2]], dim=0)
                weight_hh_reverse = weight_hh_reverse.unsqueeze(0)
                weight_ih_bi = torch.cat([weight_ih, weight_ih_reverse], dim=0)
                weight_hh_bi = torch.cat([weight_hh, weight_hh_reverse], dim=0)

                if self.normalize_weight is not None:
                    weight_ih_bi = NormalizeFunction.apply(
                        weight_ih_bi, self.normalize_weight, self.training)
                    weight_hh_bi = NormalizeFunction.apply(
                        weight_hh_bi, self.normalize_weight, self.training)

            bias_B = None
            bias_B_reverse = None
            bias_B_bi = None
            hidden_state_forward = None
            hidden_state_reverse = None
            cell_state_forward = None
            cell_state_reverse = None
            if bias_ih is not None and bias_hh is not None:
                bias_ih_chunk = bias_ih.chunk(4, 0)
                bias_ih = torch.cat(
                    [bias_ih_chunk[0], bias_ih_chunk[3], bias_ih_chunk[1], bias_ih_chunk[2]], dim=0)
                bias_hh_chunk = bias_hh.chunk(4, 0)
                bias_hh = torch.cat(
                    [bias_hh_chunk[0], bias_hh_chunk[3], bias_hh_chunk[1], bias_hh_chunk[2]], dim=0)
                bias_B = torch.cat((bias_ih, bias_hh), dim=0)
                bias_B = bias_B.unsqueeze(0)
                if self.normalize_bias is not None:
                    bias_B = NormalizeFunction.apply(
                        bias_B, self.normalize_bias, self.training)

            if self.bidirectional and bias_ih is not None and bias_hh is not None:
                bias_ih_chunk_reverse = bias_ih_reverse.chunk(4, 0)
                bias_ih_reverse = torch.cat(
                    [bias_ih_chunk_reverse[0], bias_ih_chunk_reverse[3], bias_ih_chunk_reverse[1], bias_ih_chunk_reverse[2]], dim=0)
                bias_hh_chunk_reverse = bias_hh_reverse.chunk(4, 0)
                bias_hh_reverse = torch.cat(
                    [bias_hh_chunk_reverse[0], bias_hh_chunk_reverse[3], bias_hh_chunk_reverse[1], bias_hh_chunk_reverse[2]], dim=0)
                bias_B_reverse = torch.cat(
                    (bias_ih_reverse, bias_hh_reverse), dim=0)
                bias_B_reverse = bias_B_reverse.unsqueeze(0)
                bias_B_bi = torch.cat([bias_B, bias_B_reverse], dim=0)
                if self.normalize_bias is not None:
                    bias_B_bi = NormalizeFunction.apply(
                        bias_B_bi, self.normalize_bias, self.training)

            if self.bidirectional and hx is not None:
                hidden_state_chunk = hidden_state.chunk(2, 0)
                cell_state_chunk = cell_state.chunk(2, 0)

                hidden_state_forward = hidden_state_chunk[0]
                hidden_state_reverse = hidden_state_chunk[1]
                cell_state_forward = cell_state_chunk[0]
                cell_state_reverse = cell_state_chunk[1]

            if hx is not None:
                batch_size = input.size(
                    0) if self.batch_first else input.size(1)
                seq_len = input.size(
                    1) if self.batch_first else input.size(0)
                lengths = torch.tensor([seq_len for i in range(
                    batch_size)], dtype=torch.int32, device=input.device) if lengths is None else lengths
            if lengths is not None:
                lengths = lengths.int()
            output, hy, cy = LSTMOnnxFakeFunction.apply(input, lengths, hidden_state, cell_state, weight_ih, weight_hh, bias_ih, bias_hh,
                                                        weight_ih_reverse, weight_hh_reverse, bias_ih_reverse, bias_hh_reverse,
                                                        self.input_size, self.hidden_size, self.num_layers, self.batch_first, self.dropout, self.bidirectional,
                                                        bias_B, bias_B_reverse,
                                                        hidden_state_forward, hidden_state_reverse, cell_state_forward, cell_state_reverse,
                                                        weight_ih_bi, weight_hh_bi, bias_B_bi)

            if self.normalize_data is not None:
                output = NormalizeFunction.apply(
                    output, self.normalize_data, self.training, False)

            if isinstance(orig_input, tuple):
                return (output, lengths), (hy, cy)
            else:
                return output, (hy, cy)

    def extra_repr(self):
        s = nn.GRU.extra_repr(self)
        extra_s = ',normalize_data:{normalize_data},normalize_weight:{normalize_weight},normalize_bias:{normalize_bias}'.format(
            **self.__dict__)
        return s+extra_s


__all__ = ['NormalizeFastLSTM']



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from collections import OrderedDict

import torch
import torch.nn as nn
from torch.onnx import is_in_onnx_export

from ..config import config
from ..quant import Quant, normalize_data_with_config
from ..utils import Dump, QuantMode, ScalerBuffer
from .iqtensor import (IQTensor, from_torch_tensor, platform_to_string,
                       quantlinear, torch_bmm)
from .ops import ModuleIntConfig
from .requant import Requant


class BmmIntFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, mat2, data_bits, training, momentum,
                running_x, running_y, running_o, scale_x, scale_y, scale_o,
                prefix, dump, path, mode, o_bits,
                quant, is_not_from_iqtensor1, is_not_from_iqtensor2, clamp_data):

        if training:
            saved_tensors = [input, mat2]
            # ctx.save_for_backward(input, mat2)
            ctx.bits = o_bits, data_bits
            ctx.clamp_data = clamp_data
            zero_point_x = input.zero_point if isinstance(
                input, IQTensor) else 0
            is_iq_tensor_x = True if isinstance(input, IQTensor) else False
            zero_point_y = mat2.zero_point if isinstance(
                input, IQTensor) else 0
            is_iq_tensor_y = True if isinstance(mat2, IQTensor) else False
            ctx.value = zero_point_x, is_iq_tensor_x, zero_point_y, is_iq_tensor_y
            if isinstance(input, IQTensor):
                q_input, _, max_value_x = quant.quant(
                    input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input',
                    iq_zero_point=input.zero_point)
            else:
                q_input, scale_x, max_value_x = quant.quant(
                    input.data, data_bits, mode=mode, quant_data='input')
                running_x.mul_(1-momentum).add_(momentum*max_value_x)
            if isinstance(mat2, IQTensor):
                q_mat2, _, max_value_y = quant.quant(
                    mat2.data, data_bits, scale_y, mode=QuantMode.QValue, quant_data='input', iq_zero_point=mat2.zero_point)
            else:
                q_mat2, scale_y, max_value_y = quant.quant(
                    mat2.data, data_bits, mode=mode, quant_data='input')
                running_y.mul_(1-momentum).add_(momentum*max_value_y)
            q_input = q_input.float() if data_bits <= 8 else q_input.double()
            q_mat2 = q_mat2.float() if data_bits <= 8 else q_mat2.double()
            q_outputs = torch_bmm(q_input, q_mat2)
            outputs = quant.dequant(q_outputs, scale_x * scale_y)
            out_tensor = outputs
            if o_bits is not None:
                outputs = normalize_data_with_config(outputs, clamp_data)
                out_tensor = outputs
                q_outputs, scale_o, max_value_o = quant.quant(
                    outputs, o_bits, mode=mode, quant_data='output')
                outputs = quant.dequant(q_outputs, scale_o)
                running_o.mul_(1-momentum).add_(momentum*max_value_o)
            saved_tensors += [out_tensor]
            ctx.scale = scale_x, scale_y, scale_o
            ctx.save_for_backward(*saved_tensors)

        else:
            assert running_x > 0, 'invalid running_x = 0, please finetune training before eval'
            if not isinstance(input, IQTensor):
                scale_x = ScalerBuffer(quant.running_to_scale(
                    running_x, data_bits, mode=mode))
            if not isinstance(mat2, IQTensor):
                scale_y = ScalerBuffer(quant.running_to_scale(
                    running_y, data_bits, mode=mode))
            if o_bits is not None:
                assert running_o > 0, 'invalid running_o = 0 for BmmInt'
                scale_o = ScalerBuffer(quant.running_to_scale(
                    running_o, o_bits, mode=mode))
            input_zero_point = 0
            mat2_zero_point = 0
            if isinstance(input, IQTensor):
                input_zero_point = input.zero_point
            if isinstance(mat2, IQTensor):
                mat2_zero_point = mat2.zero_point
            q_input, _, _ = quant.quant(
                input.data, data_bits, scale_x, mode=mode, quant_data='input', iq_zero_point=input_zero_point)
            q_mat2, _, _ = quant.quant(
                mat2.data, data_bits, scale_y, mode=mode, quant_data='input', iq_zero_point=mat2_zero_point)
            q_input = q_input.double()
            q_mat2 = q_mat2.double()

            q_outputs = torch_bmm(q_input, q_mat2)
            outputs = quant.dequant(q_outputs, scale_x*scale_y)
            if o_bits is not None:
                q_outputs, _, _ = quant.quant(
                    outputs, o_bits, scale_o, mode=mode, quant_data='output')
                outputs = quant.dequant(q_outputs, scale_o)

            if dump:
                name_list = ['input', 'mat2', 'outputs',
                             'q_input', 'q_mat2', 'q_outputs']
                attr_list = [input, mat2, outputs, q_input, q_mat2, q_outputs]
                Dump.dump_file(prefix, '.BmmInt.', zip(
                    name_list, attr_list), path)

        if o_bits is None:
            return outputs
        elif isinstance(scale_o, float):
            return from_torch_tensor(outputs, scale_o, o_bits)
        elif isinstance(scale_o, torch.Tensor):
            return from_torch_tensor(outputs, scale_o.item(), o_bits)
        else:
            return from_torch_tensor(outputs, scale_o.data, o_bits)

    @staticmethod
    def backward(ctx, gradoutput):
        input, mat2, output = ctx.saved_tensors
        zero_point_x, is_iq_tensor_x, zero_point_y, is_iq_tensor_y = ctx.value
        # input = input.detach().clone().requires_grad_(True)
        # mat2 = mat2.detach().clone().requires_grad_(True)
        scale_x, scale_y, scale_o = ctx.scale
        o_bits, data_bits = ctx.bits
        clamp_data = ctx.clamp_data
        grad_input = grad_mat2 = None

        if is_iq_tensor_x:
            f_input = input.data
        else:
            q_input, _, _ = Quant.quant(
                input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input')
            f_input = Quant.dequant(q_input, scale_x)
        f_input = f_input.detach().clone().requires_grad_(True)
        if is_iq_tensor_y:
            f_mat2 = mat2.data
        else:
            q_mat2, _, _ = Quant.quant(
                mat2.data, data_bits, scale_y, mode=QuantMode.QValue, quant_data='input')
            f_mat2 = Quant.dequant(q_mat2, scale_y)
        f_mat2 = f_mat2.detach().clone().requires_grad_(True)

        with torch.enable_grad():
            z = torch_bmm(f_input, f_mat2)
            if o_bits is not None:
                z = normalize_data_with_config(z, clamp_data)
            grad_input, grad_mat2 = torch.autograd.grad(
                z, (f_input, f_mat2), gradoutput)

        return grad_input, grad_mat2, None, None, None, \
            None, None, None, None,\
            None, None, None, None, None, None,\
            None, None, None, None, None,\
            None

    @staticmethod
    def symbolic(g, input, mat2,
                 data_bits, training, momentum,
                 running_x, running_y, running_o, scale_x, scale_y, scale_o,
                 prefix, dump, path, mode, o_bits,
                 quant, is_not_from_iqtensor1, is_not_from_iqtensor2, clamp_data):
        op_inner1 = None
        op_inner2 = None
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        if is_not_from_iqtensor1:
            op_inner1 = quantlinear(
                g, input, scale_x(), platform_quant, data_bits)
        if is_not_from_iqtensor2:
            op_inner2 = quantlinear(
                g, mat2, scale_y(), platform_quant, data_bits)
        param_dict = {'scale_x_f': scale_x(), 'scale_y_f': scale_y(),
                      'data_bits_i': data_bits}
        input_list = []
        if is_not_from_iqtensor1:
            input_list.append(op_inner1)
        else:
            input_list.append(input)
        if is_not_from_iqtensor2:
            input_list.append(op_inner2)
        else:
            input_list.append(mat2)
        if o_bits is not None:
            param_dict['scale_o_f'] = scale_o()
            param_dict['o_bits_i'] = o_bits
        param_dict['platform_quant_s'] = platform_quant

        return g.op("thinker::BmmInt", *input_list, **param_dict)


class BmmInt(nn.Module, ModuleIntConfig):
    def __init__(self, data_bits=8, mode=QuantMode.QValue, o_bits=None, clamp_data=None):
        nn.Module.__init__(self)
        ModuleIntConfig.__init__(
            self, data_bits=data_bits, mode=mode, o_bits=o_bits)
        self.momentum = 0.1
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.is_not_from_iqtensor1 = True
        self.is_not_from_iqtensor2 = True
        self.clamp_data = clamp_data

        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('running_y', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('scale_y', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))

    def forward(self, input, mat2):
        running_x = ScalerBuffer(self.running_x)
        running_y = ScalerBuffer(self.running_y)
        running_o = ScalerBuffer(self.running_o)
        scale_x = ScalerBuffer(self.scale_x)
        scale_y = ScalerBuffer(self.scale_y)
        scale_o = ScalerBuffer(self.scale_o)
        if isinstance(input, IQTensor):
            self.is_not_from_iqtensor1 = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)
        if isinstance(mat2, IQTensor):
            self.is_not_from_iqtensor2 = False
            if mat2.bits != self.data_bits:
                mat2 = Requant.apply(
                    mat2, mat2.bits, mat2.scale_data, self.data_bits)
            scale_y = ScalerBuffer(mat2.scale_data)
            running_y = ScalerBuffer(mat2.running_data)

        ret = BmmIntFunction.apply(input, mat2,
                                   self.data_bits, self.training, self.momentum,
                                   running_x, running_y, running_o, scale_x, scale_y, scale_o,
                                   self.prefix, self.dump, self.path, self.quant_mode, self.o_bits,
                                   self.quant, self.is_not_from_iqtensor1, self.is_not_from_iqtensor2, self.clamp_data)
        self.running_x.fill_(running_x())
        self.running_y.fill_(running_y())
        self.running_o.fill_(running_o())
        self.scale_x.fill_(scale_x())
        self.scale_y.fill_(scale_y())
        self.scale_o.fill_(scale_o())
        return ret

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        destination._metadata[prefix[:-1]
                              ] = local_metadata = dict(version=self._version)
        if is_in_onnx_export():
            assert self.running_x > 0, 'invalid running_x <=0'
            scale_x = ScalerBuffer(self.scale_x.data)
            scale_y = ScalerBuffer(self.scale_y.data)
            if self.is_not_from_iqtensor1:
                scale_x = ScalerBuffer(self.quant.running_to_scale(
                    self.running_x, self.data_bits, mode=self.quant_mode))
                self.scale_x.data.fill_(scale_x())
            if self.is_not_from_iqtensor2:
                scale_y = ScalerBuffer(self.quant.running_to_scale(
                    self.running_y, self.data_bits, mode=self.quant_mode))
                self.scale_y.data.fill_(scale_y())
            if self.o_bits is not None:
                scale_o = ScalerBuffer(self.quant.running_to_scale(
                    self.running_o, self.o_bits, mode=self.quant_mode))
                self.scale_o.data.fill_(scale_o())
        self._save_to_state_dict(destination, prefix, keep_vars)
        for name, module in self._modules.items():
            if module is not None:
                module.state_dict(destination, prefix +
                                  name + '.', keep_vars=keep_vars)
        for hook in self._state_dict_hooks.values():
            hook_result = hook(self, destination, prefix, local_metadata)
            if hook_result is not None:
                destination = hook_result
        return destination

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)



from ..ops.ops_names import (LINGER_FUNCINT_BMM_COUNTER,
                             LINGER_IQTENSOR_LAYER_COUNTER)

self_module = []


def get_current_module():
    if len(self_module) > 0:
        return self_module[-1]
    else:
        return None


def hook_pre_forward(module, input):
    setattr(module, LINGER_IQTENSOR_LAYER_COUNTER, 0)
    setattr(module, LINGER_FUNCINT_BMM_COUNTER, 0)
    self_module.append(module)


def hook_forward(module, input, output):
    cur = self_module.pop()
    assert cur == module
    setattr(module, LINGER_IQTENSOR_LAYER_COUNTER, 0)
    setattr(module, LINGER_FUNCINT_BMM_COUNTER, 0)


__all__ = ['get_current_module', 'hook_pre_forward', 'hook_forward']



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import linger
import numpy
import torch
import torch.nn as nn


def test_bmmint_net():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.linear1 = nn.Linear(12, 12)
            self.linear2 = nn.Linear(12, 12)
            self.linear3 = nn.Linear(16, 16)

        @torch.no_grad()
        def forward(self, input):

            x = self.linear1(input)

            y = self.linear2(input)

            x = x.view(8, 4, 3)

            y = y.view(8, 3, 4)
            x = torch.bmm(x, y)

            x = x.view(8, 16)
            x = self.linear3(x)

            return x
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    numpy.random.seed(1)
    # aa = torch.LongTensor([[1,2,4,5],[4,3,2,9]]).cuda()
    aa = torch.randn((8, 12), requires_grad=True).cuda()
    replace_tuple = (nn.Linear)

    net = Net().cuda()

    linger.SetFunctionBmmQuant(True)  # default  false

    net = linger.init(net, quant_modules=replace_tuple, mode=linger.QuantMode.QValue)
    criterion = nn.MSELoss()
    print(net)
    optimizer = torch.optim.Adam(net.parameters(), lr=0.01)
    loss = None
    label = torch.ones((8, 16)).cuda()
    for i in range(100):
        optimizer.zero_grad()
        out = net(aa)

    with torch.no_grad():
        net.eval()
        torch.onnx.export(net, (aa), "data.ignore/bmm_int.onnx", export_params=True, opset_version=12,
                          operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)

    linger.SetFunctionBmmQuant(False)  # set default



from setuptools import find_packages, setup

setup(
    name="pylinger",
    version="1.0.0",
    description="linger is package of fix training",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Mathematics',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Programming Language :: C++',
            'Programming Language :: Python :: 3',
    ]
)
